#!/usr/bin/env python
# coding: utf-8

"""Notice: after you edit log templates,
it is NOT reccommended to add other log messages,
especially with new log templates.
On current implementation, lt_edit will destroy
the structures of log template manager."""

from amulog import lt_common


def merge_lt(ld, ltm, ltid1, ltid2, verbose=True):
    if verbose:
        print("merge following log templates...")
        print("{0}: {1}".format(ltid1, ld.lt(ltid1)))
        print("{0}: {1}".format(ltid2, ld.lt(ltid2)))
        print()

    ltw1 = ld.lt(ltid1).ltw
    cnt1 = ld.lt(ltid1).count
    l_s = ld.lt(ltid1).lts
    ltw2 = ld.lt(ltid2).ltw
    cnt2 = ld.lt(ltid2).count
    if not len(ltw1) == len(ltw2):
        print("log template length is different, failed")
        return
    new_ltw = lt_common.merged_template(ltw1, ltw2)

    if verbose:
        print("updated log template : ltid {0}".format(ltid1))
        print("> " + " ".join(new_ltw))

        ret = input("Is it ok to update and add templates? [y/n]: ").strip().lower()
        if ret not in ("y", "yes"):
            return

    ltm.replace_lt(ltid1, new_ltw, l_s, cnt1 + cnt2)
    ltm.remove_lt(ltid2)
    ld.commit_db()
    if verbose:
        print("templates updated")
        print("updating logs of changed templates...")

    ld.db.update_log({"ltid": ltid2}, ltid=ltid1)
    ld.commit_db()
    if verbose:
        print("logs updated")
    return ltid1


def separate_lt(ld, ltm, ltid, vid, value, verbose=True):
    if verbose:
        print("separate following log template...")
        print("{0}: {1}".format(ltid, ld.lt(ltid)))
        print("add new log template where variable{0} = {1}".format(vid, value))
        print()

    l_lm = [lm for lm in ld.iter_lines(ltid=ltid)]
    l_lm_separate = []
    l_lm_remain = []
    for lm in l_lm:
        if lm.var()[vid] == value:
            l_lm_separate.append(lm)
        else:
            l_lm_remain.append(lm)

    ltw1 = lt_common.template_from_messages(l_lm_remain)
    ltw2 = lt_common.template_from_messages(l_lm_separate)
    if verbose:
        print("updated log template : ltid {0}".format(ltid))
        print("> " + " ".join(ltw1))

        print("new log template :")
        print("> " + " ".join(ltw2))

        ret = input("Is it ok to update and add templates? [y/n]: ").strip().lower()
        if ret not in ("y", "yes"):
            return

    l_s = ld.lt(ltid).lts
    ltm.replace_lt(ltid, ltw1, l_s, len(l_lm_remain))
    new_ltobj = ltm.add_lt(ltw2, l_s, len(l_lm_separate))
    if verbose:
        print("templates updated")
        print("updating logs of changed templates...")

    new_ltid = new_ltobj.ltid
    for lm in l_lm_separate:
        ld.db.update_log({"lid": lm.lid}, ltid=new_ltid)
    ld.commit_db()
    if verbose:
        print("logs updated")


def split_lt(ld, ltm, ltid, vid, verbose=True):
    if verbose:
        print("split log template {0}...".format(ltid))
        print("< " + " ".join(ld.lt(ltid).ltw))
        print("searching log instances...")

    d_lid = {}
    for lm in ld.iter_lines(ltid=ltid):
        lid = lm.lid
        w = lm.var()[vid]
        d_lid.setdefault(w, []).append(lid)

    if verbose:
        print("variable {0}: {1}".format(vid, d_lid.keys()))

    args = list(d_lid.keys())
    d_word_ltid = {}

    ltobj = ld.lt(ltid)
    l_s = ltobj.lts
    vloc = ltobj.var_location()[vid]

    # Update a template on existing ltid
    new_ltid = ltid
    new_word = args.pop(0)
    new_ltw = ltobj.ltw[:]
    new_ltw[vloc] = new_word
    count = len(d_lid[new_word])
    kwargs_updated = {"ltid": new_ltid,
                      "l_w": new_ltw,
                      "l_s": l_s,
                      "count": count}
    d_word_ltid[new_word] = new_ltid
    if verbose:
        print("updated log template : ltid {0}".format(new_ltid))
        print("> " + " ".join(new_ltw))

    # Add templates
    l_new_word = []
    l_kwargs_added = []
    while len(args) > 0:
        new_word = args.pop(0)
        new_ltw = ltobj.ltw[:]
        new_ltw[vloc] = new_word
        cnt = len(d_lid[new_word])
        kwargs = {"l_w": new_ltw,
                  "l_s": l_s,
                  "count": cnt}
        l_kwargs_added.append(kwargs)
        l_new_word.append(new_word)
        if verbose:
            print("new log template :")
            print("> " + " ".join(new_ltw))

    if verbose:
        ret = input("Is it ok to update and add templates? [y/n]: ").strip().lower()
        if ret not in ("y", "yes"):
            return

    ltm.replace_lt(**kwargs_updated)
    for kwargs, new_word in zip(l_kwargs_added, l_new_word):
        new_ltobj = ltm.add_lt(**kwargs)
        d_word_ltid[new_word] = new_ltobj.ltid

    if verbose:
        print("templates updated")
        print("updating logs of changed templates...")

    # Update log lines
    for word, l_lid in d_lid.items():
        tmp_ltid = d_word_ltid[word]
        for lid in l_lid:
            ld.db.update_log({"lid": lid}, ltid=tmp_ltid)
        if verbose:
            print("update logs of template {0}".format(tmp_ltid))

    ld.commit_db()


def fix_lt(ld, ltm, ltid, l_vid, verbose=True):
    if verbose:
        print("make variable (with no variety) into description word in following template...")
        print("log template {0}:")
        print("< " + " ".join(ld.lt(ltid).ltw))

    d_variety = {vid: set() for vid in l_vid}
    for lm in ld.iter_lines(ltid=ltid):
        for vid in l_vid:
            d_variety[vid].add(lm.var()[vid])

    d_fixed = {}
    for vid in l_vid:
        variety = d_variety[vid]
        assert len(variety) > 0
        if len(variety) == 1:
            d_fixed[vid] = variety.pop()
            if verbose:
                print("confirmed that variable {0} is stable".format(vid))
                print("fixed word : {0}".format(d_fixed[vid]))
        else:
            if verbose:
                print("variable {0} is not stable, ignored".format(vid))
                print(variety)
                print("NOTE: use lt-split to fix unstable variables")

    ltobj = ld.lt(ltid)
    new_ltw = ltobj.ltw[:]
    for vid in d_fixed.keys():
        vloc = ltobj.var_location()[vid]
        new_ltw[vloc] = d_fixed[vid]
    l_s = ltobj.lts
    cnt = ltobj.count

    if new_ltw == ltobj.ltw:
        if verbose:
            print("no changes")
        return

    if verbose:
        print("updated log template : ltid {0}".format(ltid))
        print("> " + " ".join(new_ltw))

        ret = input("Is it ok to update template? [y/n]: ").strip().lower()
        if ret not in ("y", "yes"):
            return

    ltm.replace_lt(ltid, new_ltw, l_s, cnt)
    ld.commit_db()
    if verbose:
        print("updated")
    return ltid


def free_lt(ld, ltm, ltid, l_wid, verbose=True):
    if verbose:
        print("make description word into variable (with no variety)...")
        print("log template {0}:")
        print("< " + " ".join(ld.lt(ltid).ltw))

    ltobj = ld.lt(ltid)
    new_ltw = ltobj.ltw[:]
    for wid in l_wid:
        if ltobj.ltw[wid] == lt_common.REPLACER:
            if verbose:
                print("wid {0} is not a description, ignored")
        else:
            new_ltw[wid] = lt_common.REPLACER
    l_s = ltobj.lts
    cnt = ltobj.count

    if new_ltw == ltobj.ltw:
        if verbose:
            print("no changes")
        return

    if verbose:
        print("updated log template : ltid {0}".format(ltid))
        print("> " + " ".join(new_ltw))

        ret = input("Is it ok to update template? [y/n]: ").lower()
        if ret not in ("y", "yes"):
            return

    ltm.replace_lt(ltid, new_ltw, l_s, cnt)
    ld.commit_db()
    if verbose:
        print("updated")
    return ltid


def merge_duplicated_lt(ld, ltm, verbose=True):
    from collections import defaultdict
    from itertools import combinations

    d_lt = defaultdict(list)
    for ltobj in ld.iter_lt():
        d_lt[tuple(ltobj.ltw)].append(ltobj)

    for key, l_ltobj in d_lt.items():
        if len(l_ltobj) > 1:
            if verbose:
                print("following {0} templates are duplicated.")
                for ltobj in l_ltobj:
                    print(" ".join(ltobj.ltw))

                ret = input("Is it ok to merge templates? [y/n]: ").lower()
                if ret not in ("y", "yes"):
                    print("passed as is")
                    continue
            for ltobj1, ltobj2 in combinations(l_ltobj, 2):
                merge_lt(ld, ltm, ltobj1.ltid, ltobj2.ltid,
                         verbose=False)
            print("merged")



import json


class AnonymizeMapper:

    def __init__(self, conf, online_batchsize=1000):
        self._conf = conf
        self._online_batchsize = online_batchsize
        self._d_host = None
        self._d_lt = None

        self._overwrite_method = conf["visual"]["anonymize_overwrite_method"]

    def _filepath(self):
        return self._conf["visual"]["anonymize_mapping_file"]

    def load(self):
        fp = self._filepath()
        with open(fp, mode='r', encoding='utf-8') as f:
            obj = json.load(f)
            self._d_host = obj["host_mapping"]
            self._d_lt = obj["lt_mapping"]

    def dump(self):
        fp = self._filepath()
        with open(fp, mode='wt', encoding='utf-8') as f:
            obj = {"host_mapping": self._d_host,
                   "lt_mapping": self._d_lt}
            json.dump(obj, f, indent=2)
        return fp

    def _generate_mapping(self, src_ld):
        host_mapping = {}
        host_mapping_dump = {}
        hosts = sorted(src_ld.whole_host())
        for num, host in enumerate(hosts):
            replaced_host = "host" + str(num).zfill(len(str(len(hosts))))
            host_mapping[host] = replaced_host
            host_mapping_dump[replaced_host] = host

        lt_mapping = {}
        lt_mapping_dump = {}

        from . import lt_common
        for ltobj in src_ld.iter_lt():
            new_ltw = [lt_common.REPLACER if w == lt_common.REPLACER
                       else lt_common.ANONYMIZED_DESC
                       for w in ltobj.ltw]
            new_lts = [" "] * len(ltobj.lts)
            new_lts[0] = ""
            new_lts[-1] = ""
            lt_mapping[ltobj.ltid] = {"ltw": new_ltw,
                                      "lts": new_lts}
            lt_mapping_dump[str(ltobj.ltid)] = (ltobj.ltw, ltobj.lts)

        self._host_mapping = host_mapping
        self._lt_mapping = lt_mapping
        self._d_host = host_mapping_dump
        self._d_lt = lt_mapping_dump

    def _anonymize_overwrite_legacy(self, ld):
        # overwrite hosts
        hosts = set(ld.whole_host())
        for num, host in enumerate(hosts):
            replaced_host = self._host_mapping[host]
            ld.db.update_log({"host": host}, host=replaced_host)
            ld.commit_db()

        # overwrite lt and log words
        from . import lt_common
        for ltobj in ld.iter_lt():
            new_ltw = self._lt_mapping[ltobj.ltid]["ltw"]
            new_lts = self._lt_mapping[ltobj.ltid]["lts"]
            new_ltobj = lt_common.LogTemplate(ltobj.ltid, ltobj.ltgid,
                                              new_ltw, new_lts, ltobj.count)
            ld.lttable.update_lt(new_ltobj)
            ld.db.update_lt(ltid=ltobj.ltid, ltw=new_ltw, lts=new_lts,
                            count=ltobj.count)

            ld.db.update_log({"ltid": ltobj.ltid}, l_w=new_ltw)
            ld.commit_db()

    def _anonymize_overwrite(self, ld):
        online_batchsize = ld.conf.getint("manager",
                                          "online_batchsize")

        ld.db.switch_temporal_table(ld.db.tablename_lt)
        from . import lt_common
        for ltobj in ld.iter_lt():
            new_ltobj = lt_common.LogTemplate(
                ltid=ltobj.ltid,
                ltgid=ltobj.ltgid,
                ltw=self._lt_mapping[ltobj.ltid]["ltw"],
                lts=self._lt_mapping[ltobj.ltid]["lts"],
                count=ltobj.count
            )
            ld.db.add_lt(new_ltobj)  # only update lt (w/o ltg)
            ld.commit_db()
        ld.db.apply_temporal_table(ld.db.tablename_lt)

        ld.db.switch_temporal_table(ld.db.tablename_log)
        online_counter = 0
        for lm in ld.iter_all():
            ld.add_line(lid=lm.lid,
                        ltid=lm.lt.ltid,
                        dt=lm.dt,
                        host=self._host_mapping[lm.host],
                        l_w=self._lt_mapping[lm.lt.ltid]["ltw"])
            online_counter += 1
            if online_counter >= online_batchsize:
                ld.commit_db()
                online_counter = 0
        ld.db.apply_temporal_table(ld.db.tablename_log)

    def _anonymize_migration(self, ld_src, ld_dst):
        online_batchsize = ld_dst.conf.getint("manager", "online_batchsize")

        from . import lt_common
        for ltobj in ld_src.iter_lt():
            new_ltobj = lt_common.LogTemplate(
                ltid=ltobj.ltid,
                ltgid=ltobj.ltgid,
                ltw=self._lt_mapping[ltobj.ltid]["ltw"],
                lts=self._lt_mapping[ltobj.ltid]["lts"],
                count=ltobj.count
            )
            ld_dst.db.add_lt(new_ltobj)  # only update lt (w/o ltg)
            ld_dst.commit_db()

        for ltid, ltgid in ld_src.db.iter_ltg_def():
            ld_dst.db.add_ltg(ltid, ltgid)
            ld_dst.commit_db()

        for ltid, tag in ld_src.db.iter_tag_def():
            ld_dst.db.add_tags(ltid, [tag])
            ld_dst.commit_db()

        online_counter = 0
        for lm in ld_src.iter_all():
            ld_dst.add_line(lid=lm.lid,
                            ltid=lm.lt.ltid,
                            dt=lm.dt,
                            host=self._host_mapping[lm.host],
                            l_w=self._lt_mapping[lm.lt.ltid]["ltw"])
            online_counter += 1
            if online_counter >= online_batchsize:
                ld_dst.commit_db()
                online_counter = 0

    def anonymize(self, conf_dst=None):
        from . import log_db
        if conf_dst:
            ld = log_db.LogData(self._conf, edit=False)
            ld_dst = log_db.LogData(conf_dst, edit=True, reset_db=True)
            self._generate_mapping(ld)
            self._anonymize_migration(ld, ld_dst)
        else:
            ld = log_db.LogData(self._conf, edit=True)
            self._generate_mapping(ld)
            if self._overwrite_method == "standard":
                self._anonymize_overwrite(ld)
            elif self._overwrite_method == "legacy":
                self._anonymize_overwrite_legacy(ld)
            else:
                raise ValueError

    def mapping(self):
        from . import log_db
        ld = log_db.LogData(self._conf, edit=False)
        self._generate_mapping(ld)

    def restore_host(self, host):
        return self._d_host[host]

    def restore_lt(self, ltobj):
        from . import lt_common
        return lt_common.LogTemplate(
            ltid=ltobj.ltid,
            ltgid=ltobj.ltgid,
            ltw=self._d_lt[str(ltobj.ltid)][0],
            lts=self._d_lt[str(ltobj.ltid)][1],
            count=ltobj.count
        )



#!/usr/bin/env python
# coding: utf-8

"""
Dlog: a log template generation algorithm proposed in [1].
[1] T. Li, et al. Dlog: diagnosing router events with syslogs
    for anomaly detection. The Journal of Supercomputing, pp. 1–23, 2017.

This algorithm does not work online, because it depends
on an assumption `appearance of template words are same`
and a bunch of log messages is required to check it.
"""

from collections import Counter
import logging

from amulog import lt_common

_logger = logging.getLogger(__package__)


class Node:

    def __init__(self, word, depth):
        self.word = word
        self.depth = depth
        self.count = 0
        self.childs = {}
        self.part_of_tpl = False

    def __len__(self):
        return len(self.childs)

    def __getitem__(self, key):
        if key not in self.childs:
            raise IndexError("index out of range")
        return self.childs[key]

    def add(self, num=1):
        self.count += num


class LTGenDlog(lt_common.LTGenOffline):

    def __init__(self, table, vreobj):
        super().__init__(table)
        self._vre = vreobj
        self._root = Node(None, 0)

    def _add_line(self, primary_tpl):
        words = [w for w in primary_tpl if not w == lt_common.REPLACER]
        current = self._root
        for w in words:
            if w not in current.childs:
                current.childs[w] = Node(w, current.depth + 1)
            current.childs[w].add()
            current = current.childs[w]

    @classmethod
    def _merge_subtree(cls, current_node, new_node):
        # replace a child node of the current_node into the new one
        if current_node.depth + 1 == new_node.depth:
            child = current_node.childs[new_node.word]
            for gchild_word in child.childs:
                # merge grandchild nodes
                if gchild_word in new_node.childs:
                    new_gchild = Node(gchild_word, new_node.depth + 1)
                    cls._merge_subtree(child, new_gchild)
                    cls._merge_subtree(new_node, new_gchild)
                else:
                    new_node.childs[gchild_word] = child.childs[gchild_word]
            current_node.childs[new_node.word] = new_node
        elif current_node.depth + 1 < new_node.depth:
            for child in current_node.childs.values():
                cls._merge_subtree(child, new_node)
        else:
            raise ValueError("subtree merge failure")

    @classmethod
    def _aggregate_tree(cls, node):
        d_cnt = Counter()
        # merge counts of all child nodes
        for child in node.childs.values():
            d_cnt += cls._aggregate_tree(child)

        # merge nodes of same count value
        for (word, depth), cnt in d_cnt.items():
            if cnt == node.count and cnt > 1:
                new_node = Node(word, depth)
                new_node.add(cnt)
                node.part_of_tpl = True
                new_node.part_of_tpl = True
                cls._merge_subtree(node, new_node)

        # add counts of current node
        d_cnt += Counter({(node.word, node.depth): node.count})
        return d_cnt

    def _get_primary_tpl(self, l_w):
        primary_tpl = []
        for w in l_w:
            if self._vre.match(w):
                primary_tpl.append(lt_common.REPLACER)
            else:
                primary_tpl.append(w)
        return primary_tpl

    def _restore_tpl(self, primary_tpl):
        node = self._root
        tpl = []
        for w in primary_tpl:
            if w == lt_common.REPLACER:
                tpl.append(w)
            else:
                # if w not in node.childs:
                #     import pdb; pdb.set_trace()
                node = node.childs[w]
                tpl.append(w if node.part_of_tpl else lt_common.REPLACER)
        return tpl

    def process_offline(self, d_pline):
        # make tree
        for mid, pline in d_pline.items():
            primary_tpl = self._get_primary_tpl(pline["words"])
            self._add_line(primary_tpl)

        # aggregate tree
        self._aggregate_tree(self._root)

        # restore tpl
        ret = {}
        for mid, pline in d_pline.items():
            primary_tpl = self._get_primary_tpl(pline["words"])
            tpl = self._restore_tpl(primary_tpl)
            tid, _ = self.update_table(tpl)
            ret[mid] = tid
        return ret


def init_ltgen(conf, table, **_):
    from amulog import host_alias
    from amulog.lt_regex import VariableRegex
    preprocess_fn = conf.get("log_template_dlog", "preprocess_rule")
    ha = host_alias.init_hostalias(conf)
    vreobj = VariableRegex(preprocess_fn, ha)

    return LTGenDlog(table, vreobj)



#!/usr/bin/env python
# coding: utf-8

import sys
import datetime
import logging
import argparse

from amulog import config


_logger = logging.getLogger(__package__)


def make_crf_train(ns):
    conf_load = config.open_config(ns.config_load)
    conf_dump = config.open_config(ns.config_dump)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf_dump, logger=_logger, lv=lv)

    from . import train
    from amulog import log_db
    d = parse_condition(ns.conditions)
    ld = log_db.LogData(conf_load)
    iterobj = ld.iter_lines(**d)
    print(train.crf_trainfile(conf_dump, iterobj))


def make_crf_model(ns):
    conf_load = config.open_config(ns.config_load)
    conf_dump = config.open_config(ns.config_dump)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf_dump, logger=_logger, lv=lv)

    from . import train
    output = ns.output
    if ns.train_file is None:
        from amulog import log_db
        d = parse_condition(ns.conditions)
        output_sampled = ns.output_sampled
        get_output_sampled = output_sampled is not None

        ld = log_db.LogData(conf_load)
        iterobj = [lm for lm in ld.iter_lines(**d)]
        if get_output_sampled:
            fn, l_train = train.make_crf_model(conf_dump, iterobj, output,
                                               return_sampled_messages=True)
            import pickle
            with open(output_sampled, 'wb') as f:
                pickle.dump(l_train, f)
            print("> {0}".format(output_sampled))
        else:
            fn = train.make_crf_model(conf_dump, iterobj, output)
    else:
        fn = train.make_crf_model_from_trainfile(conf_dump, ns.train_file, output)
    print("> {0}".format(fn))


def parse_condition(conditions):
    """
    Args:
        conditions (list)
    """
    d = {}
    for arg in conditions:
        if "=" not in arg:
            raise SyntaxError
        key = arg.partition("=")[0]
        if key == "ltid":
            d["ltid"] = int(arg.partition("=")[-1])
        elif key == "gid":
            d["ltgid"] = int(arg.partition("=")[-1])
        elif key == "top_date":
            date_string = arg.partition("=")[-1]
            d["top_dt"] = datetime.datetime.strptime(date_string, "%Y-%m-%d")
        elif key == "end_date":
            date_string = arg.partition("=")[-1]
            d["end_dt"] = datetime.datetime.strptime(date_string, "%Y-%m-%d")
        elif key == "date":
            date_string = arg.partition("=")[-1]
            d["top_dt"] = datetime.datetime.strptime(date_string, "%Y-%m-%d")
            d["end_dt"] = d["top_dt"] + datetime.timedelta(days=1)
        elif key == "host":
            d["host"] = arg.partition("=")[-1]
        elif key == "area":
            d["area"] = arg.partition("=")[-1]
    return d


# common argument settings
OPT_DEBUG = [["--debug"],
             {"dest": "debug", "action": "store_true",
              "help": "set logging level to debug (default: info)"}]
ARG_CONFIG_LOAD = [["config_load"],
                   {"metavar": "CONFIG_LOAD",
                    "help": "Config for db to sample training data"}]
ARG_CONFIG_DUMP = [["config_dump"],
                   {"metavar": "CONFIG_DUMP",
                    "help": "Config for crf training model"}]
# OPT_CONFIG = [["-c", "--config"],
#               {"dest": "conf_path", "metavar": "CONFIG", "action": "store",
#                # "default": config.DEFAULT_CONFIG,
#                "default": None,
#                "help": "configuration file path for amulog"}]
OPT_RECUR = [["-r", "--recur"],
             {"dest": "recur", "action": "store_true",
              "help": "recursively search files to process"}]
OPT_TRAIN_SIZE = [["-n", "--train_size"],
                  {"dest": "train_size", "action": "store",
                   "type": int, "default": 1000,
                   "help": "number of training data to sample"}]
OPT_SAMPLE_METHOD = [["-m", "--method"],
                     {"dest": "method", "action": "store",
                      "default": "all",
                      "help": "train data sampling method name. "
                              "[all, random, ltgen, leak] is available."}]
OPT_MODEL_OUTPUT = [["-o", "--output"],
                    {"dest": "output", "action": "store",
                     "default": None,
                     "help": "output model filepath"}]
ARG_DBSEARCH = [["conditions"],
                {"metavar": "CONDITION", "nargs": "+",
                 "help": ("Conditions to search log messages. "
                          "Example: MODE gid=24 date=2012-10-10 ..., "
                          "Keys: ltid, gid, date, top_date, end_date, "
                          "host, area")}]


# argument settings for each modes
# description, List[args, kwargs], func
# defined after functions because these settings use functions
DICT_ARGSET = {
    "make-crf-train": ["Generate CRF training file from DB.",
                       [OPT_DEBUG,
                        ARG_CONFIG_LOAD, ARG_CONFIG_DUMP, ARG_DBSEARCH],
                       make_crf_train],
    "make-crf-model": ["Generate CRF trained model from DB.",
                       [OPT_DEBUG, OPT_MODEL_OUTPUT,
                        ARG_CONFIG_LOAD, ARG_CONFIG_DUMP, ARG_DBSEARCH,
                        [["-t", "--train-file"],
                         {"dest": "train_file", "action": "store",
                          "default": None,
                          "help": "make model from training file"}],
                        [["-s", "--output-sampled-messages"],
                         {"dest": "output_sampled", "action": "store",
                          "default": None,
                          "help": ("output pickle of sampled LogMessage "
                                   "objects for training")}]],
                       make_crf_model],
}

USAGE_COMMANDS = "\n".join(["  {0}: {1}".format(key, val[0])
                            for key, val in sorted(DICT_ARGSET.items())])
USAGE = ("usage: {0} MODE [options and arguments] ...\n\n"
         "mode:\n".format(sys.argv[0])) + USAGE_COMMANDS


def _main():
    if len(sys.argv) < 2:
        sys.exit(USAGE)
    mode = sys.argv[1]
    if mode in ("-h", "--help"):
        sys.exit(USAGE)
    commandline = sys.argv[2:]

    desc, l_argset, func = DICT_ARGSET[mode]
    ap = argparse.ArgumentParser(prog=" ".join(sys.argv[0:2]),
                                 description=desc)
    for args, kwargs in l_argset:
        ap.add_argument(*args, **kwargs)
    ns = ap.parse_args(commandline)
    func(ns)


if __name__ == "__main__":
    _main()



#!/usr/bin/env python
# coding: utf-8

"""Matching raw log messages and its templates
that is generated by external tools."""

import re
from collections import defaultdict

# shortest match
REPLACER_REGEX_ESCAPED = re.compile(r"\\\*[A-Z]*?\\\*")


def add_esc_external(tpl):
    """Add escape sequence for imported external templates.
    It fails if the template has some statement that cannot
    be distinguished with variable replacers (e.g., *****).
    In that case, use option log_template_import.import_format_ext_esc
    and add escape sequences manually (or with another way).
    """
    from amulog import strutil
    from amulog import lt_common
    l_wild = lt_common.REPLACER_REGEX.findall(tpl)
    l_others = [strutil.add_esc(tmp)
                for tmp in lt_common.REPLACER_REGEX.split(tpl)]
    return "".join([other + wild for other, wild in zip(l_others, l_wild + [""])])


def generate_regex(tpl):
    d_name = defaultdict(list)

    def _replace_wildcard(matchobj):
        name = matchobj.group(0).strip("\\*")
        v = len(d_name[name])
        vname = "v" + name + str(v)
        d_name[name].append(vname)
        # shortest match
        regexstr = r"(?P<" + vname + r">[^*]*)"
        return regexstr

    regex_base = r"^" + re.escape(tpl) + r"$"
    tmp = REPLACER_REGEX_ESCAPED.sub(_replace_wildcard, regex_base, count=0)
    return re.compile(tmp)


def match_line(parsed_line, l_regex):
    for rid, regex in enumerate(l_regex):
        m = regex.match(parsed_line["message"])
        if m:
            return rid, m
    else:
        return None




#!/usr/bin/env python
# coding: utf-8


ONLINE_ALG = ["shiso", "drain", "lenma", "fttree"]
OFFLINE_ALG = ["va", "dlog"]
ANY_ALG = ["import", "import-ext", "re", "crf"]


def is_online(mode, alg_names, multiproc):
    if multiproc:
        return False
    elif mode == "online":
        return True
    elif mode == "offline":
        return False
    else:
        for alg_name in alg_names:
            if alg_name in OFFLINE_ALG:
                return False
        else:
            return True



import os
import re
import logging
import configparser
from abc import ABC, abstractmethod

_logger = logging.getLogger(__package__)
# DEFAULT_LABEL_CONF = "/".join((os.path.dirname(__file__),
#                                "data/lt_label.conf.sample"))


class LTTagger(ABC):

    def __init__(self):
        pass

    def get_tags(self, ltobj):
        raise NotImplementedError


class LTLabelDummy(LTTagger):

    def get_tags(self, _):
        return []


class LTLabelINI(LTTagger):
    """
    Note:
        Configuration defined earlier is prior.
    """

    group_header = "group_"
    label_header = "label_"

    def __init__(self, conf_fn, tag="group", default_label=None, default_group=None):
        from . import config
        super().__init__()
        if not os.path.exists(conf_fn):
            raise IOError("{0} not found".format(conf_fn))
        self.conf = configparser.ConfigParser()
        self.conf.read(conf_fn)
        self._tag = tag
        self.default_label = default_label
        self.default_group = default_group

        self.groups = []
        self.labels = []
        for sec in self.conf.sections():
            if sec[:len(self.group_header)] == self.group_header:
                self.groups.append(sec[len(self.group_header):])
            elif sec[:len(self.label_header)] == self.label_header:
                self.labels.append(sec[len(self.label_header):])

        self.d_group = {}  # key : group, val : [label, ...]
        self.d_rgroup = {}  # key : label, val : [group, ...]
        for group in self.groups:
            section = self.group_header + group
            for label in config.gettuple(self.conf, section, "members"):
                self.d_group.setdefault(group, []).append(label)
                self.d_rgroup.setdefault(label, []).append(group)
        self.rules = []  # [(label, (re_matchobj, ...)), ...]
        for label in self.labels:
            section = self.label_header + label
            for rulename in config.gettuple(self.conf, section, "rules"):
                l_re = []
                l_restr = config.gettuple(self.conf, section, rulename)
                for re_str in l_restr:
                    if rulename[0] == "i":
                        re_obj = re.compile(re_str, re.IGNORECASE)
                    elif rulename[0] == "c":
                        re_obj = re.compile(re_str, )
                    else:
                        raise ValueError(
                                "lt_label rulename invalid ({0})".format(
                                    rulename))
                    l_re.append(re_obj)
                self.rules.append((label, tuple(l_re)))

    @staticmethod
    def _test_rule(ltobj, t_re):
        for reobj in t_re:
            for word in ltobj.ltw:
                if reobj.match(word):
                    break
                else:
                    pass
            else:
                # no word matches
                return False
        else:
            # all reobj passes
            return True

    def get_tags(self, ltobj):
        if self._tag == "group":
            return [self.get_lt_group(ltobj)]
        elif self._tag == "label":
            return [self.get_lt_label(ltobj)]
        elif self._tag == "all":
            group = self.group_header + self.get_lt_group(ltobj)
            label = self.label_header + self.get_lt_label(ltobj)
            return [group, label]
        else:
            raise ValueError

    def get_lt_label(self, ltline):
        for label, t_re in self.rules:
            if self._test_rule(ltline, t_re):
                return label
        else:
            return self.default_label

    def get_lt_group(self, ltline):
        label = self.get_lt_label(ltline)
        return self.get_group(label)

    def get_group(self, label):
        if label is None:
            return self.default_group
        else:
            if label in self.d_rgroup:
                group = self.d_rgroup[label]
                assert len(group) == 1
                return group[0]
            else:
                return self.default_group


def init_ltlabeldummy(_):
    return LTLabelDummy()


def init_ltlabelini(conf):
    tag_file = conf.get("visual", "tag_file")
    tag_key = conf.get("visual", "tag_file_key")
    default_label = conf.get("visual", "tag_file_default_label")
    default_group = conf.get("visual", "tag_file_default_group")
    return LTLabelINI(tag_file, tag_key,
                      default_label=default_label,
                      default_group=default_group)


def generate_all_tags(conf, reset=True):
    tag_method = conf.get("visual", "tag_method")
    if tag_method == "dummy":
        tagger = init_ltlabeldummy(conf)
    elif tag_method in ("file", "ini"):
        tagger = init_ltlabelini(conf)
    else:
        raise ValueError("Invalid visual.tag_method")

    from . import log_db
    ld = log_db.LogData(conf, edit=True)
    if reset:
        ld.db.reset_tag()
    for ltobj in ld.iter_lt():
        tags = tagger.get_tags(ltobj)
        ld.db.add_tags(ltobj.ltid, tags)



#!/usr/bin/env python
# coding: utf-8

import os
import logging

import amulog.manager
from amulog import common
from amulog import strutil
from amulog import lt_common
from amulog import lt_search

_logger = logging.getLogger(__package__)


class LTGenImport(lt_common.LTGenStateless):

    def __init__(self, table, filename, mode, ltmap, lp, shuffle=False):
        super(LTGenImport, self).__init__(table)
        self._table = table
        self._d_def = common.IDDict(self._keyfunc)
        self._ltmap = lt_search.init_searcher(ltmap)
        self._lp = lp
        self._open_def(filename, mode, shuffle)

    @staticmethod
    def _keyfunc(x):
        return tuple(x)

    def _open_def(self, filename, mode, shuffle):
        if filename.strip() == "":
            errmes = "lt_import initialized with empty template set"
            _logger.info(errmes)
            return
        elif not os.path.exists(filename):
            errmes = ("log_template_import.def_path {0} is invalid".format(
                filename))
            raise IOError(errmes)

        cnt = 0
        with open(filename, 'r') as f:
            for line in f:
                if mode == "plain":
                    mes = line.rstrip("\n")
                elif mode == "manual":
                    pline = self._lp.process_message(strutil.add_esc(line))
                    mes = pline["words"]
                else:
                    raise ValueError("invalid import_mode {0}".format(
                        mode))
                if len(mes) == 0:
                    continue
                ltw, lts = self._lp.process_statement(mes)
                self.add_definition(ltw)
                cnt += 1
        _logger.info("LTGenImport: {0} template imported".format(cnt))
        if shuffle:
            self._ltmap.shuffle()

    def add_definition(self, ltw):
        defid = self._d_def.add(ltw)
        self._ltmap.add(defid, ltw)
        _logger.debug("defid {0}: {1}".format(defid, ltw))

    def update_definition(self, old_ltw, new_ltw):
        defid = self._ltmap.remove(old_ltw)
        _logger.debug("defid remove {0}: {1}".format(defid, old_ltw))
        self._d_def.set_item(defid, new_ltw)
        self._ltmap.add(defid, new_ltw)
        _logger.debug("defid {0}: {1}".format(defid, new_ltw))

    def generate_tpl(self, pline):
        defid = self._ltmap.search(pline["words"])
        if defid is None:
            return None
        else:
            tpl = self._d_def.get(defid)
            return tpl


def init_ltgen_import(conf, table, shuffle=False, **_):
    fn = conf.get("log_template_import", "def_path")
    mode = conf.get("log_template_import", "import_format")
    ltmap = conf.get("log_template_import", "search_method")
    from amulog import log_db
    lp = amulog.manager.load_log2seq(conf)
    return LTGenImport(table, fn, mode, ltmap, lp, shuffle)



import logging
import pandas as pd

from . import strutil
from . import lt_common
from amsemantics import TopicClustering

_logger = logging.getLogger(__package__)


class LTGroupSemantics(lt_common.LTGroupOffline):

    def __init__(self, lttable, normalizer,
                 lda_modelname="gensim",
                 lda_knowledge_sources=None,
                 cache_dir="/tmp",
                 rfc_use_cache=True, rfc_document_unit="rfc",
                 random_seed=None,
                 stop_words=None,
                 use_nltk_stopwords=False,
                 use_sklearn_stopwords=False,
                 lda_n_topics=None, cluster_eps=None,
                 cluster_size_min=5,
                 tuning_metrics="cluster"):
        super().__init__(lttable)
        self._lognorm = normalizer
        self._lda_modelname = lda_modelname
        self._lda_knowledge_sources = lda_knowledge_sources
        self._cache_dir = cache_dir
        self._rfc_use_cache = rfc_use_cache
        self._rfc_document_unit = rfc_document_unit
        self._stop_words = stop_words
        self._use_nltk_stopwords = use_nltk_stopwords,
        self._use_sklearn_stopwords = use_sklearn_stopwords,
        self._random_seed = random_seed
        self._lda_n_topics = lda_n_topics
        self._cluster_eps = cluster_eps
        self._cluster_size_min = cluster_size_min
        self._tuning_metrics = tuning_metrics

        self._sc = None
        self._tuning_rules = None
        self._l_ltid = None
        self._training_docs = None
        self._input_docs = None

    def _tokenize(self, ltobj):
        return self._lognorm.process_line(
            [strutil.restore_esc(w) for w in ltobj.desc()]
        )

    def set_tuning_rules(self, *args, **kwargs):
        self._tuning_rules = (args, kwargs)

    def set_documents(self, verbose=False):
        self._l_ltid = [ltobj.ltid for ltobj in self.lttable]
        self._input_docs = [self._tokenize(ltobj)
                            for ltobj in self.lttable]
        self._training_docs = self.get_training_data(
            self._input_docs, verbose=verbose
        )

    def make(self, verbose=False):
        if self._input_docs is None:
            self.set_documents(verbose=verbose)

        self._sc = TopicClustering(
            model=self._lda_modelname,
            redistribute=True,
            random_seed=self._random_seed,
            stop_words=self._stop_words,
            use_nltk_stopwords=self._use_nltk_stopwords,
            use_sklearn_stopwords=self._use_sklearn_stopwords,
            lda_n_topics=self._lda_n_topics,
            cluster_eps=self._cluster_eps,
            cluster_size_min=self._cluster_size_min,
            tuning_metrics=self._tuning_metrics,
            verbose=verbose
        )
        if self._tuning_rules is not None:
            args, kwargs = self._tuning_rules
            self._sc.set_tuning_rules(*args, **kwargs)
        clusters = self._sc.fit(self._input_docs, training_docs=self._training_docs)

        self._n_groups = len(clusters)
        for cls, l_idx in clusters.items():
            for idx in l_idx:
                ltid = self._l_ltid[idx]
                self.add_lt(int(cls), self.lttable[ltid])
        return self.update_lttable(self.lttable)

    def get_inspection(self, topn=10):
        assert self._sc is not None
        return self._sc.inspection(topn=topn)

    def get_training_data(self, docs, verbose=False):
        """
        Generate training data based on the knowledge_sources parameter.

        Args:
            docs (List[List[str]]): Input documents.
            verbose (boolean)
        """
        if self._lda_knowledge_sources is None:
            knowledge_sources = ["self"]
        else:
            knowledge_sources = self._lda_knowledge_sources

        training_docs = []
        for source_name in knowledge_sources:
            if source_name == "self":
                training_docs += docs
            elif source_name == "rfc":
                from amsemantics.source import rfcdoc
                kwargs = {"cache_dir": self._cache_dir,
                          "use_cache": self._rfc_use_cache,
                          "normalizer": self._lognorm}
                if self._rfc_document_unit == "rfc":
                    loader = rfcdoc.RFCLoader(**kwargs)
                elif self._rfc_document_unit == "section":
                    loader = rfcdoc.RFCSectionLoader(**kwargs)
                elif self._rfc_document_unit == "line":
                    loader = rfcdoc.RFCLinesLoader(**kwargs)
                else:
                    raise ValueError("invalid rfc_document_unit")
                for rfc in loader.iter_all():
                    try:
                        if verbose:
                            print("loading RFC {0}".format(rfc.n))
                        docs = loader.get_document(rfc)
                        training_docs += docs
                    except FileNotFoundError:
                        mes = "skip RFC {0}: line file not found".format(rfc.n)
                        if verbose:
                            print(mes)
                        _logger.warning(mes)

        return training_docs

    def lda_search_param(self, verbose=False):
        if self._training_docs is None:
            self.set_documents(verbose)
        results = [(n, p, c) for n, p, c
                   in self._sc.lda_search_param(self._training_docs)]
        return pd.DataFrame(results,
                            columns=["n_topics", "log_perplexity", "coherence"])

    def get_visual_data(self, dim=2, with_mean=True, with_std=True):
        return self._sc.get_principal_components(
            input_docs=self._input_docs,
            n_components=dim,
            with_mean=with_mean, with_std=with_std
        )

    def show_pyldavis(self, mds="pcoa"):
        return self._sc.show_pyldavis(mds=mds)


class LTGroupSemanticsMultiData(LTGroupSemantics):

    def __init__(self, lttables, normalizer, **kwargs):
        super().__init__(lttables[0], normalizer, **kwargs)
        self._l_lttable = lttables

        # used afeter make
        self._l_ltid = None
        self._l_domain = None

    def set_documents(self, verbose=False):
        self._l_ltid = [ltobj.ltid for ltobj in self.lttable]
        self._l_domain = []
        self._input_docs = []
        for domain, lttable in enumerate(self._l_lttable):
            for ltobj in lttable:
                self._l_ltid.append(ltobj.ltid)
                self._l_domain.append(domain)
                self._input_docs.append(self._tokenize(ltobj))
        self._training_docs = self.get_training_data(
            self._input_docs, verbose=verbose
        )

    def make(self, verbose=False):
        if self._input_docs is None:
            self.set_documents(verbose=verbose)

        self._sc = TopicClustering(
            model=self._lda_modelname,
            redistribute=True,
            stop_words=self._stop_words,
            random_seed=self._random_seed,
            lda_n_topics=self._lda_n_topics,
            cluster_eps=self._cluster_eps,
            cluster_size_min=self._cluster_size_min,
            tuning_metrics=self._tuning_metrics,
            verbose=verbose
        )
        if self._tuning_rules is not None:
            args, kwargs = self._tuning_rules
            self._sc.set_tuning_rules(*args, **kwargs)
        clusters = self._sc.fit(self._input_docs, training_docs=self._training_docs)

        self._n_groups = len(clusters)
        for cls, l_idx in clusters.items():
            for idx in l_idx:
                ltid = self._l_ltid[idx]
                domain = self._l_domain[idx]
                self._l_lttable[domain][ltid].ltgid = cls

        return self._l_lttable

    def get_labels(self):
        import numpy as np
        return np.array(self._l_domain), np.array(self._l_ltid), self._sc.cluster_labels


def init_nlp_normalizer(conf):
    from . import config
    from amsemantics import Normalizer
    filters = config.getlist(conf, "nlp_preprocess", "filters")
    mreplacer_sources = config.getlist(conf, "nlp_preprocess",
                                       "replacer_sources")
    vreplacer_source = conf["nlp_preprocess"]["variable_rule"]
    ha_source = conf["manager"]["host_alias_filename"]
    if ha_source == "":
        ha_source = None
    lemma_exception = config.getlist(conf, "nlp_preprocess",
                                     "lemma_exception")
    th_word_length = conf.getint("nlp_preprocess",
                                 "remove_short_word_length")
    return Normalizer(filters, mreplacer_sources,
                      vreplacer_source, ha_source,
                      lemma_exception, th_word_length)


def _get_ltgroup_semantic_params(conf):
    from . import config

    lda_model = conf["log_template_group_semantics"]["lda_model"]
    lda_knowledge_sources = config.getlist(
        conf, "log_template_group_semantics", "lda_knowledge_sources"
    )

    cache_dir = conf["general"]["cache_dir"]
    rfc_use_cache = conf["log_template_group_semantics"]["rfc_use_cache"]
    rfc_document_unit = conf["log_template_group_semantics"]["rfc_document_unit"]
    stop_words = config.getlist(conf, "log_template_group_semantics",
                                "lda_stop_words")
    use_nltk_stopwords = conf["log_template_group_semantics"]["lda_use_nltk_stopwords"]
    use_sklearn_stopwords = conf["log_template_group_semantics"]["lda_use_sklearn_stopwords"]
    random_seed_str = conf["log_template_group_semantics"]["lda_seed"]
    if random_seed_str.isdigit():
        random_seed = int(random_seed_str)
    else:
        random_seed = None
    lda_n_topics_str = conf["log_template_group_semantics"]["lda_n_topics"]
    if lda_n_topics_str == "":
        lda_n_topics = None
    else:
        lda_n_topics = int(lda_n_topics_str)
    cluster_eps_str = conf["log_template_group_semantics"]["cluster_eps"]
    if cluster_eps_str == "":
        cluster_eps = None
    else:
        cluster_eps = float(cluster_eps_str)
    cluster_size_min = conf.getint("log_template_group_semantics",
                                   "cluster_size_min")
    tuning_metrics = conf["log_template_group_semantics"]["tuning_metrics"]

    return {
        "lda_modelname": lda_model,
        "lda_knowledge_sources": lda_knowledge_sources,
        "cache_dir": cache_dir,
        "rfc_use_cache": rfc_use_cache,
        "rfc_document_unit": rfc_document_unit,
        "random_seed": random_seed,
        "stop_words": stop_words,
        "use_nltk_stopwords": use_nltk_stopwords,
        "use_sklearn_stopwords": use_sklearn_stopwords,
        "lda_n_topics": lda_n_topics,
        "cluster_eps": cluster_eps,
        "cluster_size_min": cluster_size_min,
        "tuning_metrics": tuning_metrics,
    }


def init_ltgroup_semantics(conf, lttable):
    from . import config
    normalizer = init_nlp_normalizer(conf)
    kwargs = _get_ltgroup_semantic_params(conf)
    ltgroup = LTGroupSemantics(lttable, normalizer, **kwargs)

    if kwargs["lda_n_topics"] is None or kwargs["cluster_eps"] is None:
        union_rules_str = config.getlist(conf,
                                         "log_template_group_semantics",
                                         "tuning_union_rules")
        union_rules = [rulestr.split("|")
                       for rulestr in union_rules_str]
        separation_rules_str = config.getlist(conf,
                                              "log_template_group_semantics",
                                              "tuning_separation_rules")
        separation_rules = [rulestr.split("|")
                            for rulestr in separation_rules_str]
        if len(union_rules) > 0 or len(separation_rules) > 0:
            term_class = conf["log_template_group_semantics"]["tuning_term_class"]
            tuning_topn = conf.getint("log_template_group_semantics", "tuning_topwords")
            ltgroup.set_tuning_rules(union_rules, separation_rules,
                                     term_class=term_class, topn=tuning_topn)

    return ltgroup


def init_ltgroup_semantics_multi_domain(l_conf, l_lttable):
    from . import config
    conf = l_conf[0]
    normalizer = init_nlp_normalizer(conf)
    kwargs = _get_ltgroup_semantic_params(conf)
    ltgroup = LTGroupSemanticsMultiData(l_lttable, normalizer, **kwargs,)

    if kwargs["lda_n_topics"] is None or kwargs["cluster_eps"] is None:
        union_rules_str = config.getlist(conf,
                                         "log_template_group_semantics",
                                         "tuning_union_rules")
        union_rules = [rulestr.split("|")
                       for rulestr in union_rules_str]
        separation_rules_str = config.getlist(conf,
                                              "log_template_group_semantics",
                                              "tuning_separation_rules")
        separation_rules = [rulestr.split("|")
                            for rulestr in separation_rules_str]
        if len(union_rules) > 0 or len(separation_rules) > 0:
            term_class = conf["log_template_group_semantics"]["tuning_term_class"]
            tuning_topn = conf.getint("log_template_group_semantics", "tuning_topwords")
            ltgroup.set_tuning_rules(union_rules, separation_rules,
                                     term_class=term_class, topn=tuning_topn)

    return ltgroup



#!/usr/bin/env python
# coding: utf-8

import logging

from amulog import cli
from amulog import common
from amulog import config

_logger = logging.getLogger(__package__)


def show_lt_words(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)

    from amulog import log_db
    ld = log_db.LogData(conf)

    from . import search
    d = search.agg_words(ld, target="all")
    print(common.cli_table(sorted(d.items(), key=lambda x: x[1],
                                  reverse=True)))


def show_lt_descriptions(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)

    from amulog import log_db
    ld = log_db.LogData(conf)

    from . import search
    d = search.agg_words(ld, target="description")
    print(common.cli_table(sorted(d.items(), key=lambda x: x[1],
                                  reverse=True)))


def show_lt_variables(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)

    from amulog import log_db
    ld = log_db.LogData(conf)

    from . import search
    d = search.agg_words(ld, target="variable")
    if ns.repld:
        import re
        reobj = re.compile(r"[0-9]+")
        keys = list(d.keys())
        for k in keys:
            new_k = reobj.sub(r"\d", k)
            if k == new_k:
                pass
            else:
                d[new_k] += d.pop(k)

    print(common.cli_table(sorted(d.items(), key=lambda x: x[1],
                                  reverse=True)))


def show_lt_breakdown(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)

    from . import search
    ltid = ns.ltid
    limit = ns.lines
    print(search.breakdown_lt(conf, ltid, limit))


def show_lt_vstable(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)

    from amulog import log_db
    ld = log_db.LogData(conf)

    from . import search
    for ret in search.stable_variables(ld, ltid=ns.ltid, th=1):
        print(ld.show_lt_info(ret["ltid"]))
        msg = "variable {0.vid} (word location {0.vloc}): {0.dict}".format(ret)
        print("  " + msg)


def show_lt_positions(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)

    from amulog import log_db
    ld = log_db.LogData(conf)
    ltobj = ld.lt(ns.ltid)

    from amulog import lt_common
    line_lt = ["lt:"] + ltobj.ltw
    line_wid = ["wid:"] + list(range(len(ltobj.ltw)))
    line_vid = ["vid:", ]
    vcnt = 0
    for w in ltobj.ltw:
        if w == lt_common.REPLACER:
            line_vid.append(str(vcnt))
            vcnt += 1
        else:
            line_vid.append("")
    print(common.cli_table([line_lt, line_wid, line_vid]))


def lttool_merge(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)

    from amulog import log_db
    from amulog import manager
    ld = log_db.LogData(conf)
    ltm = manager.init_manager(ld)
    verbose = (not ns.force)

    from . import lt_tool
    lt_tool.merge_lt(ld, ltm, ns.ltid1, ns.ltid2, verbose)


def lttool_separate(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)

    from amulog import log_db
    from amulog import manager
    ld = log_db.LogData(conf)
    ltm = manager.init_manager(ld)
    verbose = (not ns.force)

    from . import lt_tool
    lt_tool.separate_lt(ld, ltm, ns.ltid, ns.vid, ns.word, verbose)


def lttool_split(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)

    from amulog import log_db
    from amulog import manager
    ld = log_db.LogData(conf)
    ltm = manager.init_manager(ld)
    verbose = (not ns.force)

    from . import lt_tool
    lt_tool.split_lt(ld, ltm, ns.ltid, ns.vid, verbose)


def lttool_fix(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)

    from amulog import log_db
    from amulog import manager
    ld = log_db.LogData(conf)
    ltm = manager.init_manager(ld)
    verbose = (not ns.force)

    from . import lt_tool
    lt_tool.fix_lt(ld, ltm, ns.ltid, ns.vids, verbose)


def lttool_free(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)

    from amulog import log_db
    from amulog import manager
    ld = log_db.LogData(conf)
    ltm = manager.init_manager(ld)
    verbose = (not ns.force)

    from . import lt_tool
    lt_tool.free_lt(ld, ltm, ns.ltid, ns.wids, verbose)


def lttool_merge_dup(ns):
    conf = config.open_config(ns.conf_path)
    lv = logging.DEBUG if ns.debug else logging.INFO
    config.set_common_logging(conf, logger=_logger, lv=lv)

    from amulog import log_db
    from amulog import manager
    ld = log_db.LogData(conf)
    ltm = manager.init_manager(ld)
    verbose = (not ns.force)

    from . import lt_tool
    lt_tool.merge_duplicated_lt(ld, ltm, verbose)


# common argument settings
OPT_DEBUG = [["--debug"],
             {"dest": "debug", "action": "store_true",
              "help": "set logging level to debug (default: info)"}]
OPT_CONFIG = [["-c", "--config"],
              {"dest": "conf_path", "metavar": "CONFIG", "action": "store",
               "default": None,
               "help": "configuration file path for amulog"}]
OPT_FORCE = [["-f", "--force"],
             {"dest": "force", "action": "store_true",
              "help": "no confirmation nor verbose notification"}]

# argument settings for each modes
# description, List[args, kwargs], func
# defined after functions because these settings use functions
DICT_ARGSET = {
    "show-lt-words": ["Show words and their counts in all messages",
                      [OPT_CONFIG, OPT_DEBUG],
                      show_lt_words],
    "show-lt-description": ["Show description words and their counts",
                            [OPT_CONFIG, OPT_DEBUG],
                            show_lt_descriptions],
    "show-lt-variable": ["Show variable words and their counts",
                         [OPT_CONFIG, OPT_DEBUG,
                          [["-d", "--digit"],
                           {"dest": "repld", "action": "store_true",
                            "help": "replace digit to \\d"}]],
                         show_lt_variables],
    "show-lt-breakdown": ["Show variable samples in a log template.",
                          [OPT_CONFIG, OPT_DEBUG,
                           [["-l", "--ltid"],
                            {"dest": "ltid", "metavar": "LTID",
                             "action": "store", "type": int, "default": None,
                             "help": "log template identifier to investigate"}],
                           [["-n", "--number"],
                            {"dest": "lines", "metavar": "LINES",
                             "action": "store", "type": int, "default": 5,
                             "help": "number of variable candidates to show"}]],
                          show_lt_breakdown],
    "show-lt-vstable": ["Show stable variables in the template.",
                        [OPT_CONFIG, OPT_DEBUG,
                         [["-l", "--ltid"],
                          {"dest": "ltid", "metavar": "LTID",
                           "action": "store", "type": int, "default": None,
                           "help": "log template identifier to investigate"}],
                         [["-n", "--number"],
                          {"dest": "number", "metavar": "NUMBER",
                           "action": "store", "type": int, "default": 1,
                           "help": "thureshold number to be stable"}],
                         ],
                        show_lt_vstable],
    "show-lt-positions": [("Show wid and vid in the template.@ "
                           "Use to confirm positional arguments for edit commands."),
                          [OPT_CONFIG, OPT_DEBUG,
                           [["ltid"],
                            {"metavar": "LTID", "action": "store", "type": int,
                             "help": "log template identifier"}]],
                          show_lt_positions],
    "lt-merge": ["Merge 2 templates and generate a new template.",
                 [OPT_CONFIG, OPT_DEBUG, OPT_FORCE,
                  [["ltid1"],
                   {"metavar": "LTID1", "action": "store", "type": int,
                    "help": "first log template to merge"}],
                  [["ltid2"],
                   {"metavar": "LTID2", "action": "store", "type": int,
                    "help": "second log template to merge"}], ],
                 lttool_merge],
    "lt-separate": [("Separate messages satisfying the given condition "
                     "and make it a new log template."),
                    [OPT_CONFIG, OPT_DEBUG, OPT_FORCE,
                     [["ltid"],
                      {"metavar": "LTID", "action": "store", "type": int,
                       "help": "log template indentifier"}],
                     [["vid"],
                      {"metavar": "VARIABLE-ID", "action": "store",
                       "type": int,
                       "help": "variable identifier to fix"}],
                     [["word"],
                      {"metavar": "WORD", "action": "store",
                       "help": "a word to fix in the location of vid"}]],
                    lttool_separate],
    "lt-split": [("Fix variables as descriptions to split template clusters.@ "
                  "Use carefully because this function may cause "
                  "generating enormous unexpected templates."),
                 [OPT_CONFIG, OPT_DEBUG, OPT_FORCE,
                  [["ltid"],
                   {"metavar": "LTID", "action": "store", "type": int,
                    "help": "log template indentifier"}],
                  [["vid"],
                   {"metavar": "VARIABLE-ID", "action": "store",
                    "help": "variable identifier to fix"}]],
                 lttool_split],
    "lt-fix": [("Fix specified stable variable as a description.@ "
                "Usually this command add one new template."
                "For unstable variables, use lt-split or lt-separate."),
               [OPT_CONFIG, OPT_DEBUG, OPT_FORCE,
                [["ltid"],
                 {"metavar": "LTID", "action": "store", "type": int,
                  "help": "log template to fix"}],
                [["vids"],
                 {"metavar": "VARIABLE-IDs", "nargs": "+", "type": int,
                  "help": "variable identifiers to fix"}]],
               lttool_fix],
    "lt-free": ["Make a description word into a variable.",
                [OPT_CONFIG, OPT_DEBUG, OPT_FORCE,
                 [["ltid"],
                  {"metavar": "LTID", "action": "store", "type": int,
                   "help": "log template indentifier"}],
                 [["wids"],
                  {"metavar": "WORD-IDs", "nargs": "+", "type": int,
                   "help": "word locations to fix"}]],
                lttool_free],
    "lt-merge-dup": ["Merge templates with same structure.",
                     [OPT_CONFIG, OPT_DEBUG, OPT_FORCE],
                     lttool_merge_dup],
}


def main():
    cli.main(DICT_ARGSET)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# coding: utf-8


from collections import defaultdict


class RegexTable(object):

    def __init__(self, l_tpl, l_regex):
        assert len(l_tpl) == len(l_regex)
        self._table = self._make_table(l_tpl, l_regex)

    def _make_table(self, l_tpl, l_regex):
        return [(tplid, reobj) for tplid, reobj in enumerate(l_regex)]

    def search(self, mes):
        for tplid, reobj in self._table:
            matchobj = reobj.match(mes)
            if matchobj:
                return tplid, matchobj
        else:
            return None

    def shuffle(self):
        import random
        random.shuffle(self._table)


class RegexHashTable(RegexTable):

    def __init__(self, l_tpl, l_regex, headlen=5):
        assert isinstance(headlen, int)
        assert headlen > 0
        self._headlen = headlen
        self._table = None
        super().__init__(l_tpl, l_regex)

    def _make_table(self, l_tpl, l_regex):
        table = defaultdict(list)
        for tplid, (tpl, reobj) in enumerate(zip(l_tpl, l_regex)):
            if self._head_isstable(tpl, self._headlen):
                key = tpl[:self._headlen]
            else:
                key = ""
            table[key].append((tplid, reobj))
        return table

    @staticmethod
    def _head_isstable(tpl, headlen):
        from amulog import lt_common
        replacer = lt_common.REPLACER_REGEX
        matchobj = replacer.search(tpl)
        if matchobj:
            return matchobj.start() >= headlen
        else:
            return True

    def search(self, mes):
        key = mes[:self._headlen]
        if key not in self._table:
            key = ""

        for tplid, reobj in self._table[key]:
            matchobj = reobj.match(mes)
            if matchobj:
                return tplid, matchobj
        for tplid, reobj in self._table[""]:
            matchobj = reobj.match(mes)
            if matchobj:
                return tplid, matchobj
        return None

    def shuffle(self):
        import random
        for key, subtable in self._table.items():
            random.shuffle(subtable)



#!/usr/bin/env python
# coding: utf-8

"""Convert Sequencial-string training data into ItemSequence data
by feature generation based on feature definitions.
The features are defined by 3 components.
- crf feature template: defining positional relations of each features
- crf label definition: defining mid-labels for feature "pos"
- input data (Sequencial-string)
"""

import os
import re

import pycrfsuite


class FeatureExtracter:
    DEFAULT_TEMPLATE = "/".join((os.path.dirname(os.path.abspath(__file__)),
                                 "../../data/crf_template"))
    TPL_PARSER = re.compile(r"^([a-z0-9]+)(\[[-0-9]+\])?$")
    TPL_FIELD = {"0": 0, "w": 0, "word": 0,
                 "1": 1, "pos": 1, "mid": 1,
                 "2": 2, "y": 2, "label": 2,
                 "bos": None, "eos": None,
                 "F": None}

    def __init__(self, template_fp=None, ig_key=[], ig_weight=0.1):
        if template_fp is None or template_fp == "":
            self._template = self._load_template(self.DEFAULT_TEMPLATE)
        else:
            self._template = self._load_template(template_fp)
        self._ig_key = ig_key  # unknown keys
        self._ig_val = ig_weight  # weight for unknown keys
        self._bos = False
        self._eos = False

    def _load_template_line(self, line):
        l_rule = []
        line = line.rstrip()
        if line == "":
            return None

        if ":" in line:
            temp = line.split(":")
            line = temp[0].strip()
            weight = float(temp[1].strip())
        else:
            weight = 1.0
        l_cond = [s.strip() for s in line.split(",")]
        name = "|".join(l_cond)
        for cond in l_cond:
            m = self.TPL_PARSER.match(cond)
            if m is None:
                raise SyntaxError(
                    "Invalid syntax of feature template ({0})".format(
                        line))
            field, offset_str = m.groups()
            if offset_str is None:
                offset = None
            else:
                offset = int(offset_str.strip("[").rstrip("]"))
            l_rule.append(tuple([field, offset]))
        return name, tuple(l_rule), weight

    def _load_template(self, fp):
        template = []  # name, tuples(field (int), offset (int))), weight
        with open(fp, 'r') as f:
            for line in f:
                ret = self._load_template_line(line)
                if ret is not None:
                    template.append(tuple(ret))
        return tuple(template)

    def _get_item(self, item, field):
        if field in self.TPL_FIELD:
            return item[self.TPL_FIELD[field]]
        else:
            raise NotImplementedError

    def feature(self, l_items):
        l_items_range = range(len(l_items))
        ret = []
        for wid, item in enumerate(l_items):
            d_feature = {}
            for name, l_rule, weight in self._template:
                subfeature = []
                for field, offset in l_rule:
                    if field == "bos":
                        if wid == 0:
                            subfeature.append("__BOS__")
                        else:
                            subfeature = None
                            break
                    elif field == "eos":
                        if wid == len(l_items) - 1:
                            subfeature.append("__EOS__")
                        else:
                            subfeature = None
                            break
                    else:
                        p = wid + offset
                        if p in l_items_range:
                            temp_item = l_items[p]
                            val = self._get_item(temp_item, field)
                            subfeature.append(val)
                        else:
                            subfeature = None
                            break
                if subfeature is not None:
                    s = set(subfeature)
                    if len(s) == 1:
                        if len(s & set(self._ig_key)) == 1:
                            weight = weight * self._ig_val
                    fval = "|".join(subfeature)
                    key = "=".join((name, fval))
                    d_feature[key] = weight
            ret.append(d_feature)

        return pycrfsuite.ItemSequence(ret)

    @staticmethod
    def label(l_items):
        return [item[-1] for item in l_items]



# !/usr/bin/env python
# coding: utf-8

import logging

from . import lt_common

_logger = logging.getLogger(__package__)


class LTGroupFuzzyHash(lt_common.LTGroup):
    """Classify templates based on ssdeep, an implementation of fuzzy hashing.
    Fuzzy hashing evaluate the similarity of string data. This class compares
    strings of templates including symbols and variable replacements, and
    make groups that have large scores of fuzzy hash comparizon each other.
    
    """

    def __init__(self, lttable, th=1, mem_hash=True):
        super(LTGroupFuzzyHash, self).__init__()
        self.th = th
        self._lttable = lttable
        self._mem_hash = mem_hash
        self._d_hash = {}

    def add(self, lt_new):
        l_score = self._calc_score(lt_new)
        if len(l_score) == 0:
            gid = self._next_groupid()
        else:
            ltid, score = max(l_score, key=lambda x: x[1])
            if score >= self.th:
                gid = self._d_rgroup[ltid]
            else:
                gid = self._next_groupid()
        self.add_lt(gid, lt_new)
        return gid

    def _calc_score(self, lt_new):
        try:
            import ssdeep
        except ImportError:
            raise ImportError(
                "ltgroup algorithm <ssdeep> needs python package ssdeep")
        ret = []
        h1 = ssdeep.hash(str(lt_new))
        if self._mem_hash:
            if len(self._d_hash) == 0:
                # initialize d_hash
                for lt in self._lttable:
                    h = ssdeep.hash(str(lt))
                    self._d_hash[lt.ltid] = h
            for ltid, lt_temp in enumerate(self._lttable):
                h2 = self._d_hash[lt_temp.ltid]
                score = ssdeep.compare(h1, h2)
                ret.append((ltid, score))
            self._d_hash[lt_new.ltid] = h1
        else:
            for lt_temp in self._lttable:
                ltid = lt_temp.ltid
                score = ssdeep.hash_score(str(lt_new), str(lt_temp))
                ret.append((ltid, score))
        return ret


def edit_distance(m1, m2):
    """Calculate Levenshtein edit distance of 2 tokenized messages.
    This function considers wildcards."""

    table = [[0] * (len(m2) + 1) for _ in range(len(m1) + 1)]

    for i in range(len(m1) + 1):
        table[i][0] = i

    for j in range(len(m2) + 1):
        table[0][j] = j

    for i in range(1, len(m1) + 1):
        for j in range(1, len(m2) + 1):
            if m1[i - 1] == m2[j - 1] or \
                    m1[i - 1] == lt_common.REPLACER or \
                    m2[j - 1] == lt_common.REPLACER:
                cost = 0
            else:
                cost = 1
            table[i][j] = min(table[i - 1][j] + 1, table[i][j - 1] + 1,
                              table[i - 1][j - 1] + cost)
    return table[-1][-1]


def norm_edit_distance(m1, m2):
    return edit_distance(m1, m2) / max(len(m1), len(m2))


#def generate_lt(conf, targets, check_import=False):
#    from . import strutil
#    from . import log_db
#    lp = log_db.load_log2seq(conf)
#    table = lt_common.TemplateTable()
#    ltgen_import = lt_common.init_ltgen_methods(conf, table, "import")
#    ltgen = lt_common.init_ltgen_methods(conf, table)
#    s_tpl = set()
#
#    def _is_known(pline, ltgen_import, flag):
#        if not flag:
#            return False
#        else:
#            tid, dummy = ltgen_import.process_line(pline)
#            return tid is not None
#
#    for fp in targets:
#        _logger.info("processing {0} start".format(fp))
#        with open(fp, 'r') as f:
#            for line in f:
#                pline = lp.process_line(strutil.add_esc(line))
#                if _is_known(pline, ltgen_import, check_import):
#                    # recorded in imported definition, ignore
#                    pass
#                else:
#                    tpl = ltgen.estimate_tpl(pline["words"], pline["symbols"])
#                    s_tpl.add(tuple(tpl))
#
#        _logger.info("processing {0} done".format(fp))
#    return s_tpl



import os
import datetime
import random
import re
import configparser
import numpy as np

from amulog import config

DEFAULT_CONFIG = "/".join((os.path.dirname(os.path.abspath(__file__)),
                           "./data/testlog.conf"))


class TestLogGenerator:
    _var_re = re.compile(r"\$.+?\$")

    def __init__(self, conf_fn, seed=None):
        if seed is None:
            random.seed()
        else:
            random.seed(seed)

        self.conf = configparser.ConfigParser()
        self.conf.read(conf_fn)
        self.term = config.getterm(self.conf, "main", "term")
        self.top_dt, self.end_dt = self.term
        self.d_host = {}
        for group in config.gettuple(self.conf, "main", "host_groups"):
            for host in config.gettuple(self.conf, "main", "group_" + group):
                self.d_host.setdefault(group, []).append(host)

        self.l_event = []
        for event_name in config.gettuple(self.conf, "main", "events"):
            self._generate_event(event_name)

        self.l_log = []
        for event in self.l_event:
            self._generate_log(event)

    def _dt_rand(self, top_dt, end_dt):
        return self._dt_delta_rand(top_dt,
                                   datetime.timedelta(0), end_dt - top_dt)

    @staticmethod
    def _dt_delta_rand(top_dt, durmin, durmax):
        deltasec = 24 * 60 * 60 * (durmax.days - durmin.days) + \
                   durmax.seconds - durmin.seconds
        seconds = datetime.timedelta(seconds=random.randint(0, deltasec - 1))
        return top_dt + durmin + seconds

    @staticmethod
    def rand_uniform(top_dt, end_dt, lambd):
        """Generate a random event that follows uniform distribution of
        LAMBD times a day on the average.

        Args:
            top_dt (datetime.datetime): The start time of generated event.
            end_dt (datetime.datetime): The end time of generated event.
            lambd (float): The average of appearance of generated event per a day.

        Returns:
            list[datetime.datetime]
        """
        ret = []
        total_dt = (end_dt - top_dt)
        avtimes = 1.0 * lambd * total_dt.total_seconds() / (24 * 60 * 60)
        times = int(np.random.poisson(avtimes))
        for i in range(times):
            deltasec = int(total_dt.total_seconds() * random.random())
            dt = top_dt + datetime.timedelta(seconds=deltasec)
            ret.append(dt)
        return ret

    @classmethod
    def _rand_exp(cls, top_dt, end_dt, lambd):
        """Generate a random event that follows Poisson process.
        The event interval matches exponential distribution of
        LAMBD times a day on the average.

        Args:
            top_dt (datetime.datetime): The start time of generated event.
            end_dt (datetime.datetime): The end time of generated event.
            lambd (float): The average of appearance of generated event per a day.

        Yields:
            datetime.datetime
        """
        temp_dt = top_dt
        temp_dt = cls._rand_next_exp(temp_dt, lambd)
        while temp_dt < end_dt:
            yield temp_dt
            temp_dt = cls._rand_next_exp(temp_dt, lambd)

    @staticmethod
    def _rand_next_exp(dt, lambd):
        return dt + datetime.timedelta(seconds=1) * int(
            24 * 60 * 60 * random.expovariate(lambd))

    def _generate_event(self, event_name):

        section = "event_" + event_name

        def _recur(dt, host, event_name):
            if not self.conf.has_option(section, "recurrence"):
                return
            if self.conf.getboolean(section, "recurrence"):
                if random.random() < self.conf.getfloat(section, "recur_p"):
                    durmin = config.getdur(self.conf, section, "recur_dur_min")
                    durmax = config.getdur(self.conf, section, "recur_dur_max")
                    new_dt = self._dt_delta_rand(dt, durmin, durmax)
                    _add_event(new_dt, host, event_name)

        def _add_event(dt, host, event_name):
            info = {}
            if self.conf.has_option(section, "info"):
                for i in config.gettuple(self.conf, section, "info"):
                    if i == "ifname":
                        info[i] = random.choice(
                            config.gettuple(self.conf, section, "ifname"))
                    elif i == "user":
                        info[i] = random.choice(
                            config.gettuple(self.conf, section, "user"))
            self.l_event.append((dt, host, event_name, info))
            _recur(dt, host, event_name)

        for group in config.gettuple(self.conf, section, "groups"):
            for host in self.d_host[group]:
                occ = self.conf.get(section, "occurrence")
                if occ == "random_uniform":
                    freq = self.conf.getfloat(section, "frequency")
                    for dt in self.rand_uniform(self.top_dt, self.end_dt,
                                                freq):
                        _add_event(dt, host, event_name)
                elif occ in ("random", "random_exp"):
                    freq = self.conf.getfloat(section, "frequency")
                    for dt in self._rand_exp(self.top_dt, self.end_dt, freq):
                        _add_event(dt, host, event_name)
                elif occ == "hourly":
                    dursec = 60 * 60
                    first_dt = self.top_dt + datetime.timedelta(
                        seconds=random.randint(0, dursec - 1))
                    now_dt = first_dt
                    while now_dt < self.end_dt:
                        _add_event(now_dt, host, event_name)
                        now_dt += datetime.timedelta(seconds=dursec)
                elif occ == "daily":
                    dursec = 24 * 60 * 60
                    first_dt = self.top_dt + datetime.timedelta(
                        seconds=random.randint(0, dursec - 1))
                    now_dt = first_dt
                    while now_dt < self.end_dt:
                        _add_event(now_dt, host, event_name)
                        now_dt += datetime.timedelta(seconds=dursec)

    def _generate_log(self, event):
        dt = event[0]
        host = event[1]
        event_name = event[2]
        info = event[3]
        for log_name in config.gettuple(self.conf,
                                        "event_" + event_name, "logs"):
            section = "log_" + log_name
            mode = self.conf.get(section, "mode")
            form = self.conf.get(section, "format")

            mes = form
            while True:
                match = self._var_re.search(mes)
                if match is None:
                    break
                var_type = match.group().strip("$")
                if var_type in info.keys():
                    var_string = info[var_type]
                elif var_type == "pid":
                    var_string = str(random.randint(1, 65535))
                elif var_type == "host":
                    var_string = host
                else:
                    raise ValueError
                mes = "".join((mes[:match.start()] + var_string +
                               mes[match.end():]))

            if mode == "each":
                self.l_log.append((dt, host, mes))
            elif mode == "delay_rand":
                delay_min = config.getdur(self.conf, section, "delay_min")
                delay_max = config.getdur(self.conf, section, "delay_max")
                log_dt = self._dt_delta_rand(dt, delay_min, delay_max)
                self.l_log.append((log_dt, host, mes))
            elif mode == "drop_rand":
                drop_p = self.conf.getfloat(section, "drop_p")
                if random.random() > drop_p:
                    self.l_log.append((dt, host, mes))
            elif mode == "other_host_rand":
                l_host = []
                for t_group in config.gettuple(self.conf, section, "groups"):
                    for t_host in self.d_host[t_group]:
                        if not t_host == host:
                            l_host.append(t_host)
                self.l_log.append((dt, random.choice(l_host), mes))

    def dump_log(self, output):
        l_line = sorted(self.l_log, key=lambda x: x[0])
        if output is None:
            for line in l_line:
                print(" ".join((line[0].strftime("%Y-%m-%d %H:%M:%S"),
                                line[1], line[2])))
        else:
            with open(output, 'w') as f:
                for line in l_line:
                    f.write(" ".join((line[0].strftime("%Y-%m-%d %H:%M:%S"),
                                      line[1], line[2])) + "\n")


def generate_testdata(fn=None, output=None, seed=None):
    if fn is None:
        fn = DEFAULT_CONFIG
    tlg = TestLogGenerator(fn, seed)
    tlg.dump_log(output)



#!/usr/bin/env python
# coding: utf-8

import os
import logging
import pycrfsuite

from amulog import lt_common
from amulog.alg.crf import _items, _convert

_logger = logging.getLogger(__package__)
DEFAULT_TEMPLATE = "/".join((os.path.dirname(os.path.abspath(__file__)),
                             "../../data/crf_template.default"))


class LTGenCRF(lt_common.LTGenStateless):
    LABEL_DESC = "D"
    LABEL_VAR = "V"
    LABEL_DUMMY = "N"

    POS_UNKNOWN = "unknown"

    def __init__(self, table, model, verbose, feature_template, unknown_weight,
                 vreobj, normalizer_conf=None, normalizer_rule="DEFAULT"):
        super().__init__(table)
        self._model = model
        self._verbose = verbose
        self._feature_template = feature_template

        self._trainer = None
        self._tagger = None
        self._converter = _convert.FeatureExtracter(self._feature_template,
                                                    [self.POS_UNKNOWN],
                                                    unknown_weight)
        self._vre = vreobj
        self._norm = normalizer_conf is not None  # use normalization or not
        if self._norm:
            self._norm_rule = normalizer_rule
            self._init_normalizer(normalizer_conf)

    def middle_label(self, w):
        if self._vre is None:
            return w
        else:
            return self._vre.label(w)

    def _init_normalizer(self, normalize_conf):
        from log_normalizer import LogNormalizer
        self._ln = LogNormalizer(normalize_conf)

    def _normalize_line(self, l_w, rule_name="DEFAULT"):
        words = []
        length_vec = []
        for w in l_w:
            try:
                if w == "" or w == lt_common.REPLACER:
                    replaced = w
                else:
                    replaced = self._ln.replace_word(w, rule_name)
            except IndexError:
                import pdb; pdb.set_trace()
            else:
                words.extend(replaced)
                length_vec.append(len(replaced))
        return words, length_vec

    @classmethod
    def _tpl_labels(cls, org_words, tpl):
        labels = []
        for org_w, tpl_w in zip(org_words, tpl):
            if org_w == tpl_w:
                labels.append(cls.LABEL_DESC)
            elif tpl_w == lt_common.REPLACER:
                labels.append(cls.LABEL_VAR)
            else:
                raise ValueError("template structure invalud")
        return labels

    @classmethod
    def _normalized_labels(cls, org_words, tpl, length_vec):
        labels = []
        for org_w, tpl_w, length in zip(org_words, tpl, length_vec):
            if org_w == tpl_w:
                labels.extend([cls.LABEL_DESC] * length)
            elif tpl_w == lt_common.REPLACER:
                labels.extend([cls.LABEL_VAR] * length)
            else:
                raise ValueError("template structure invalud")
        return labels

    @classmethod
    def _restore_tpl(cls, org_words, labels):
        tpl = []
        for w, label in zip(org_words, labels):
            if label == cls.LABEL_DESC:
                tpl.append(w)
            elif label == cls.LABEL_VAR:
                tpl.append(lt_common.REPLACER)
            elif label == cls.LABEL_DUMMY:
                raise ValueError("Some word labeled as DUMMY in LTGenCRF")
            else:
                raise ValueError("Unknown labels in LTGenCRF")
        return tpl

    @classmethod
    def _restore_normalized_tpl(cls, org_words, labels, length_vec):
        assert len(labels) == sum(length_vec)
        import numpy as np
        org_tpl = []
        # slicer: iterate slice range of labels
        slicer = zip(np.cumsum([0] + length_vec[:-1]), np.cumsum(length_vec))
        for org_wid, (start, stop) in enumerate(slicer):
            tmp_labels = labels[start, stop]
            if cls.LABEL_VAR in tmp_labels:
                org_tpl.append(lt_common.REPLACER)
            elif cls.LABEL_DUMMY in tmp_labels:
                raise ValueError("Some word labeled as DUMMY in LTGenCRF")
            else:
                org_tpl.append(org_words[org_wid])
        return org_tpl

    def trainitems(self, lm):
        if self._norm:
            words, length_vec = self._normalize_line(lm.l_w, self._norm_rule)
            labels = self._normalized_labels(lm.l_w, lm.lt.ltw, length_vec)
        else:
            words = lm.l_w
            labels = self._tpl_labels(lm.l_w, lm.lt.ltw)

        assert len(words) == len(labels)
        mlabels = [self.middle_label(w) for w in words]
        return list(zip(words, mlabels, labels))

    def init_trainer(self, alg="lbfgs", verbose=False):
        self._trainer = pycrfsuite.Trainer(verbose=verbose)
        self._trainer.select(alg, "crf1d")
        d = {}  # for parameter tuning, edit this
        if len(d) > 0:
            self._trainer.set_params(d)

    def _train_model(self, iterable_items, model):
        if model is None:
            model = self._model
        for lineitems in iterable_items:
            xseq = self._converter.feature(lineitems)
            yseq = self._converter.label(lineitems)
            self._trainer.append(xseq, yseq)
        self._trainer.train(model)

    def train(self, iterable_lm, output=None):
        def _iter_items():
            for lm in iterable_lm:
                yield self.trainitems(lm)

        if output is None:
            output = self._model
        self._train_model(_iter_items(), output)
        return output

    def train_from_file(self, fp, output=None):
        self._train_model(_items.load_trainitems(fp), output)
        return output

    def init_tagger(self):
        if not os.path.exists(self._model):
            raise IOError("No trained CRF model for LTGenCRF")
        self._tagger = pycrfsuite.Tagger()
        self._tagger.open(self._model)

    def close_tagger(self):
        if self._tagger is not None:
            self._tagger.close()

    def _tag_line(self, l_w):
        lineitems = _items.line2items(l_w, midlabel_func=self.middle_label,
                                      dummy_label=self.LABEL_DUMMY)
        if self._tagger is None:
            self.init_tagger()
        fset = self._converter.feature(lineitems)
        l_label = self._tagger.tag(fset)
        return l_label

    def generate_tpl(self, pline):
        org_words = pline["words"]
        if self._norm:
            l_w, length_vec = self._normalize_line(org_words, self._norm_rule)
            l_label = self._tag_line(l_w)
            tpl = self._restore_normalized_tpl(org_words, l_label, length_vec)
        else:
            l_label = self._tag_line(org_words)
            tpl = self._restore_tpl(org_words, l_label)
        return tpl


# class MeasureAccuracy():
#
#    def __init__(self, conf, s_ltid=None):
#        """
#        Args:
#            conf
#            s_ltid (set): if not None, use limited messages
#                          of given ltid as train data"""
#        from . import log_db
#        self.ld = log_db.LogData(conf)
#        self.ld.init_ltmanager()
#        self.conf = conf
#        self.s_ltid = s_ltid
#        self.measure_lt_method = conf.get("measure_lt", "lt_method")
#        self.sample_from = conf.get("measure_lt", "sample_from")
#        self.sample_rules = self._rules(
#            conf.get("measure_lt", "sample_rules"))
#        self.trials = conf.getint("measure_lt", "train_trials")
#        self.results = []
#
#        if self.sample_from == "cross":
#            self.cross_k = conf.getint("measure_lt", "cross_k")
#            assert self.trials <= self.cross_k, "trials is larger than k"
#            self._eval_cross()
#        elif self.sample_from == "diff":
#            self.sample_train_rules = self._rules(
#                conf.get("measure_lt", "sample_train_rules"))
#            self.train_sample_method = conf.get("measure_lt",
#                                                "train_sample_method")
#            self.train_size = conf.getint("measure_lt", "train_size")
#            self._eval_diff()
#        elif self.sample_from == "file":
#            self.fn = conf.get("measure_lt", "sample_from_file")
#            self.train_sample_method = conf.get("measure_lt",
#                                                "train_sample_method")
#            self.train_size = conf.getint("measure_lt", "train_size")
#            self._eval_file()
#        else:
#            raise NotImplementedError
#
#    @staticmethod
#    def _rules(rulestr):
#        ret = {}
#        l_rule = [rule.strip() for rule in rulestr.split(",")]
#        for rule in l_rule:
#            if rule == "":
#                continue
#            key, val = [v.strip() for v in rule.split("=")]
#            if key in ("host", "area"):
#                ret[key] = val
#            elif key == "top_date":
#                assert not "top_dt" in ret
#                ret["top_dt"] = datetime.datetime.strptime(val, "%Y-%m-%d")
#            elif key == "top_dt":
#                assert not "top_dt" in ret
#                ret["top_dt"] = datetime.datetime.strptime(val,
#                                                           "%Y-%m-%d %H:%M:%S")
#            elif key == "end_date":
#                assert not "end_dt" in ret
#                ret["end_dt"] = datetime.datetime.strptime(
#                    val, "%Y-%m-%d") + datetime.timedelta(days=1)
#            elif key == "end_dt":
#                assert not "end_dt" in ret
#                ret["end_dt"] = datetime.datetime.strptime(val,
#                                                           "%Y-%m-%d %H:%M:%S")
#            else:
#                raise NotImplementedError
#        return ret
#
#    @staticmethod
#    def _make_ltidmap(l_ltid):
#        ltidmap = defaultdict(int)
#        for ltid in l_ltid:
#            ltidmap[ltid] += 1
#        return ltidmap
#
#    def _eval_cross(self):
#
#        def divide_size(size, groups):
#            base = int(size // groups)
#            ret = [base] * groups
#            surplus = size - (base * groups)
#            for i in range(surplus):
#                ret[i] += 1
#            assert sum(ret) == size
#            return ret
#
#        def agg_dict(l_d):
#            keyset = set()
#            for d in l_d:
#                keyset.update(d.keys())
#
#            new_d = defaultdict(int)
#            for ltid in keyset:
#                for d in l_d:
#                    new_d[ltid] += d[ltid]
#            return new_d
#
#        l_labeled = []  # (ltid, train)
#        for line in self.ld.iter_lines(**self.sample_rules):
#            l_labeled.append((line.lt.ltid, items.lm2trainitems(line)))
#        random.shuffle(l_labeled)
#
#        l_group = []
#        l_group_ltidlist = []
#        basenum = 0
#        for size in divide_size(len(l_labeled), self.cross_k):
#            l_ltid, l_lm = zip(*l_labeled[basenum:basenum + size])
#            l_group.append(l_lm)
#            l_group_ltidlist.append(l_ltid)
#            basenum += size
#        assert sum([len(g) for g in l_group]) == len(l_labeled)
#        del l_labeled
#
#        for trial in range(self.trials):
#            l_train = None
#            l_test = []
#            l_test_ltidlist = []
#            for gid, group, ltidlist in zip(range(self.cross_k),
#                                            l_group, l_group_ltidlist):
#                if gid == trial:
#                    assert l_train is None
#                    l_train = group
#                    train_ltidlist = ltidlist
#                else:
#                    l_test += group
#                    test_ltidlist += ltidlist
#            # test_ltidmap = agg_dict(l_test_ltidlist)
#
#            l_train, train_ltidlist = self._filter_train(l_train,
#                                                         train_ltidlist)
#            d_result = self._trial(l_train, l_test,
#                                   train_ltidlist, test_ltidlist)
#            self.results.append(d_result)
#
#    def _eval_diff(self):
#
#        l_test, test_ltidlist = self._load_test_diff()
#
#        l_train_all = []  # (ltid, lineitems)
#        for line in self.ld.iter_lines(**self.sample_train_rules):
#            l_train_all.append(line)
#
#        if self.train_sample_method == "all":
#            if self.trials > 1:
#                raise UserWarning(("measure_lt.trials is ignored "
#                                   "because the results must be static"))
#            train_ltidlist = [lm.lt.ltid for lm in l_train_all]
#            l_train = [items.lm2trainitems(lm) for lm in l_train_all]
#            l_train, train_ltidlist = self._filter_train(l_train,
#                                                         train_ltidlist)
#            d_result = self._trial(l_train, l_test,
#                                   train_ltidlist, test_ltidlist)
#            self.results.append(d_result)
#        elif self.train_sample_method == "random":
#            for i in range(self.trials):
#                l_train, train_ltidlist = train_sample_random(l_train_all,
#                                                              self.train_size)
#                l_train, train_ltidlist = self._filter_train(l_train,
#                                                             train_ltidlist)
#                d_result = self._trial(l_train, l_test,
#                                       train_ltidlist, test_ltidlist)
#                self.results.append(d_result)
#        elif self.train_sample_method == "random-va":
#            ltgen_va, ret_va = va_preprocess(self.conf, l_train_all)
#            for i in range(self.trials):
#                l_train, train_ltidlist = train_sample_random_va(
#                    l_train_all, self.train_size, ltgen_va, ret_va)
#                l_train, train_ltidlist = self._filter_train(l_train,
#                                                             train_ltidlist)
#                d_result = self._trial(l_train, l_test,
#                                       train_ltidlist, test_ltidlist)
#                self.results.append(d_result)
#        else:
#            raise NotImplementedError
#
#    def _eval_file(self):
#        l_train, train_ltidlist = self._load_train_file()
#        l_test, test_ltidlist = self._load_test_diff()
#        l_train, train_ltidlist = self._filter_train(l_train,
#                                                     train_ltidlist)
#        d_result = self._trial(l_train, l_test,
#                               train_ltidlist, test_ltidlist)
#        self.results.append(d_result)
#
#    def _load_test_diff(self):
#        l_test = []
#        test_ltidlist = []
#        test_ltidmap = defaultdict(int)
#        for line in self.ld.iter_lines(**self.sample_rules):
#            l_test.append(items.lm2trainitems(line))
#            test_ltidlist.append(line.lt.ltid)
#        return l_test, test_ltidlist
#
#    def _load_train_file(self):
#        l_test = []
#        for lineitems in items.iter_items_from_file(self.fn):
#            l_test.append(lineitems)
#        return l_test, []
#
#    def _filter_train(self, l_train, train_ltidlist):
#        if self.s_ltid is None:
#            return l_train, train_ltidlist
#        ret_l_train = []
#        ret_train_ltidlist = []
#        for trainobj, ltid in zip(l_train, train_ltidlist):
#            if ltid in self.s_ltid:
#                ret_l_train.append(trainobj)
#                ret_train_ltidlist.append(ltid)
#        assert len(ret_l_train) > 0
#        _logger.info("train data decreased with ltid-based filter "
#                     "({0} -> {1})".format(len(l_train), len(ret_l_train)))
#        return ret_l_train, ret_train_ltidlist
#
#    def _trial(self, l_train, l_test, l_ltid_train, l_ltid_test):
#
#        def form_template(ltgen, l_w, l_label):
#            tpl = []
#            for w, label in zip(l_w, l_label):
#                if label == ltgen.LABEL_DESC:
#                    tpl.append(w)
#                elif label == ltgen.LABEL_VAR:
#                    tpl.append(ltgen._sym)
#                elif label == ltgen.LABEL_DUMMY:
#                    raise ValueError("Some word labeled as DUMMY in LTGenCRF")
#                else:
#                    raise ValueError
#            return tpl
#
#        table = self.ld.ltm._table
#        ltgen = lt_common.init_ltgen_methods(self.conf, table, "crf")
#        d_ltid_test = self._make_ltidmap(l_ltid_test)
#
#        ltgen.init_trainer()
#        ltgen.train(l_train)
#
#        wa_numer = 0.0
#        wa_denom = 0.0
#        la_numer = 0.0
#        la_denom = 0.0
#        ta_numer = 0.0
#        ta_denom = 0.0
#        d_failure = defaultdict(int)
#
#        for lineitems, ltid in zip(l_test, l_ltid_test):
#            l_correct = items.items2label(lineitems)
#            l_w = [item[0] for item in lineitems]
#            l_label_correct = [item[-1] for item in lineitems]
#            l_label = ltgen.label_line(lineitems)
#
#            for wid, (w_correct, w_label) in enumerate(zip(l_correct,
#                                                           l_label)):
#                wa_denom += 1
#                if w_correct == w_label:
#                    wa_numer += 1
#                else:
#                    d_failure[(ltid, wid)] += 1
#            if ltid is None:
#                tpl = form_template(ltgen, l_w, l_label_correct)
#                assert ltgen._table.exists(tpl)
#                ltid = ltgen._table.get_tid(tpl)
#            cnt = d_ltid_test[ltid]
#            la_denom += 1
#            ta_denom += 1.0 / cnt
#            if l_correct == l_label:
#                la_numer += 1
#                ta_numer += 1.0 / cnt
#
#        d_result = {"word_acc": wa_numer / wa_denom,
#                    "line_acc": la_numer / la_denom,
#                    "tpl_acc": ta_numer / ta_denom,
#                    "train_size": len(l_train),
#                    "test_size": len(l_test),
#                    "train_tpl_size": len(set(l_ltid_train)),
#                    "test_tpl_size": len(set(l_ltid_test)),
#                    "dict_ltid": d_ltid_test,
#                    "failure": d_failure}
#        return d_result
#
#    def info(self):
#        buf = []
#        buf.append(("# Experiment for measuring "
#                    "log template generation accuracy"))
#        if self.sample_from == "cross":
#            buf.append("# type: Cross-validation (k = {0})".format(
#                self.cross_k))
#            buf.append("# data-source: db({0})".format(self.sample_rules))
#        elif self.sample_from in ("diff", "file"):
#            buf.append("# type: Experiment with different data range / domain")
#            if self.sample_from == "diff":
#                buf.append("# train-source: db({0})".format(
#                    self.sample_train_rules))
#            elif self.sample_from == "file":
#                buf.append("# train-source: file({0})".format(self.fn))
#            buf.append("# test-source: db({0})".format(self.sample_rules))
#        buf.append("# trials: {0}".format(self.trials))
#        return "\n".join(buf)
#
#    def result(self):
#        import numpy as np
#        buf = []
#        for rid, result in enumerate(self.results):
#            buf.append("Experiment {0}".format(rid))
#            for key, val in result.items():
#                if not type(val) in (list, tuple, dict, defaultdict):
#                    buf.append("{0} {1}".format(key, val))
#            buf.append("")
#
#        buf.append("# General result")
#        arr_wa = np.array([d["word_acc"] for d in self.results])
#        wa = np.average(arr_wa)
#        wa_err = np.std(arr_wa) / np.sqrt(arr_wa.size)
#        buf.append("Average word accuracy: {0} (err: {1})".format(wa, wa_err))
#
#        arr_la = np.array([d["line_acc"] for d in self.results])
#        la = np.average(arr_la)
#        la_err = np.std(arr_la) / np.sqrt(arr_la.size)
#        buf.append("Average line accuracy: {0} (err: {1})".format(la, la_err))
#
#        arr_ta = np.array([d["tpl_acc"] for d in self.results])
#        ta = np.average(arr_ta)
#        ta_err = np.std(arr_ta) / np.sqrt(arr_ta.size)
#        buf.append("Average template accuracy: {0} (err: {1})".format(
#            ta, ta_err))
#
#        return "\n".join(buf)
#
#    def failure_report(self, ld=None):
#
#        def _failure_place(ld, ltid, wid):
#            if ld is None:
#                return None
#            else:
#                tpl = []
#                for temp_wid, w in enumerate(ld.lt(ltid).ltw):
#                    if temp_wid == wid:
#                        tpl.append("<{0}>".format(w))
#                    else:
#                        tpl.append(w)
#                return " ".join(tpl)
#
#        buf = []
#        buf2 = []
#
#        whole_keys = set()
#        for result in self.results:
#            whole_keys = whole_keys | set(result["failure"].keys())
#
#        d_average = defaultdict(float)
#        for result in self.results:
#            d_fail = result["failure"]
#            for key in whole_keys:
#                if key in d_fail:
#                    d_average[key] += 1.0 * d_fail[key] / len(self.results)
#
#        d_ltid = self.results[0]["dict_ltid"]
#        for key, cnt in sorted(d_average.items(), key=lambda x: x[1],
#                               reverse=True):
#            ltid, wid = key
#            ratio = 1.0 * cnt / d_ltid[ltid]
#            buf.append([ltid, wid, ":", int(cnt), "({0})".format(ratio)])
#            tpl_info = _failure_place(ld, ltid, wid)
#            if tpl_info is not None:
#                buf2.append(tpl_info)
#
#        if len(buf2) == 0:
#            return "\n".join(buf)
#        else:
#            from . import common
#            ret = []
#            for buf_line, buf2_line in zip(common.cli_table(buf).split("\n"),
#                                           buf2):
#                ret.append(buf_line)
#                ret.append(buf2_line)
#                ret.append("")
#            return "\n".join(ret)


def init_ltgen_crf(conf, table, **_):
    model = conf.get("log_template_crf", "model_filename")
    verbose = conf.getboolean("log_template_crf", "verbose")
    feature_template = conf.get("log_template_crf", "feature_template")
    if feature_template.strip() == "":
        feature_template = DEFAULT_TEMPLATE
    unknown_weight = conf.getfloat("log_template_crf", "unknown_key_weight")
    normalizer_conf = conf.get("log_template_crf", "normalizer_conf")
    if normalizer_conf.strip() == "":
        normalizer_conf = None
    normalizer_rule = conf.get("log_template_crf", "normalizer_rule")

    middle_fn = conf.get("log_template_crf", "middle_label_rule")
    if middle_fn.strip() == "":
        vreobj = None
    else:
        from amulog import host_alias
        from amulog import lt_regex
        ha = host_alias.init_hostalias(conf)
        vreobj = lt_regex.VariableRegex(middle_fn, ha, LTGenCRF.POS_UNKNOWN)

    return LTGenCRF(table, model, verbose, feature_template, unknown_weight,
                    vreobj, normalizer_conf, normalizer_rule)



#!/usr/bin/env python
# coding: utf-8

"""Modify templates that is generated by external tools.
amulog can import templates that is splitted as a sequence of words,
but many external tools do not provide them in such style.
This script redifine templates by templates
and their corresponding log messages.
"""

import numpy as np

from amulog import lt_common


def _get_variable_mask(mes, matchobj):
    vm = np.array([False] * len(mes))
    d = matchobj.groupdict()
    for k in d:
        vm[matchobj.start(k):matchobj.end(k)] = True
    return vm


def _get_word_span(l_tpl_w, l_tpl_s):
    ws = []
    place = len(l_tpl_s[0])
    for w, s in zip(l_tpl_w, l_tpl_s[1:]):
        place_next = place + len(w)
        ws.append((place, place_next))
        place_next += len(s)
        place = place_next
    return ws


def redefine_tpl(tpl, parsed_line, matchobj=None):
    mes = parsed_line["message"]
    l_tpl_w = parsed_line["words"]
    l_tpl_s = parsed_line["symbols"]

    if matchobj is None:
        from . import tpl_match
        tpl_regex = tpl_match.generate_regex(tpl)
        matchobj = tpl_regex.match(mes)
        if matchobj is None:
            errmes = "message {0} not matched to tpl {1}".format(mes, tpl)
            raise ValueError(errmes)

    vm = _get_variable_mask(mes, matchobj)
    ws = _get_word_span(l_tpl_w, l_tpl_s)

    new_tpl = []
    for w, span in zip(l_tpl_w, ws):
        if True in vm[span[0]:span[1]]:
            new_tpl.append(lt_common.REPLACER)
        else:
            new_tpl.append(w)
    return new_tpl



#!/usr/bin/env python
# coding: utf-8

"""
FT-tree: a log template generation algorithm proposed in [1].
[1] S. Zhang, et al. Syslog Processing for Switch Failure Diagnosis and
    Prediction in Datacenter Networks. In IEEE/ACM 25th International
    Symposium on Quality of Service (IWQoS), pp. 1–10, 2017.

see also https://github.com/WeibinMeng/ft-tree
"""


import logging
from collections import defaultdict

from amulog import lt_common

_logger = logging.getLogger(__package__)


class Node:

    def __init__(self, word, depth):
        self.word = word
        self.depth = depth
        self.childs = {}

    def __len__(self):
        return len(self.childs)


class LTGenFTTree(lt_common.LTGen):

    def __init__(self, table, max_child=6, cut_depth=3, message_type_func=None):
        super().__init__(table)
        self._max_child = max_child
        self._cut_depth = cut_depth
        self._d_words = defaultdict(int)
        self._tree = {}

        if message_type_func is None:
            self._type_func = self.message_type_none
        else:
            self._type_func = message_type_func

    def load(self, loadobj):
        self._tree, self._d_words = loadobj

    def dumpobj(self):
        return self._tree, self._d_words

    @staticmethod
    def message_type_none(words):
        return None, words

    @staticmethod
    def message_type_top(words):
        return words[0], words[1:]

    @staticmethod
    def message_type_length(words):
        return str(len(words)), words

    def _add_line(self, pline):
        list_words = [(w, self._d_words[w]) for w in pline["words"]]
        message_type, list_words = self._type_func(list_words)
        list_words.sort(key=lambda x: (x[1], x[0]), reverse=True)

        if message_type not in self._tree:
            self._tree[message_type] = Node(message_type, 0)
        current = self._tree[message_type]

        s_tpl = set()
        for w, _ in list_words:
            if current.childs is None:
                break
            elif w not in current.childs:
                current.childs[w] = Node(w, current.depth + 1)
                # pruning
                if len(current.childs) >= self._max_child and \
                        current.depth >= self._cut_depth:
                    current.childs = None
                    break
            s_tpl.add(w)
            current = current.childs[w]

        tpl = [w if w in s_tpl else lt_common.REPLACER
               for w in pline["words"]]
        return tpl

    def process_offline(self, d_pline):
        # make word dictionary
        for pline in d_pline.values():
            for w in pline["words"]:
                self._d_words[w] += 1

        # make tree
        ret = {}
        for mid, pline in d_pline.items():
            tpl = self._add_line(pline)
            ret[mid] = self.update_table(tpl)
        return ret

    def process_line(self, pline):
        # update word dictionary
        for w in pline["words"]:
            self._d_words[w] += 1

        # update tree
        tpl = self._add_line(pline)
        return self.update_table(tpl)

    def generate_tpl(self, pline):
        tid, _ = self.process_line(pline)
        return self._table[tid]


def init_ltgen(conf, table, **_):
    max_child = conf.getint("log_template_fttree", "max_child")
    cut_depth = conf.getint("log_template_fttree", "cut_depth")
    type_func_name = conf.get("log_template_fttree", "type_func")
    if type_func_name == "top":
        message_type_func = LTGenFTTree.message_type_top
    elif type_func_name == "length":
        message_type_func = LTGenFTTree.message_type_length
    else:
        message_type_func = None
    return LTGenFTTree(table, max_child, cut_depth, message_type_func)


def get_param_candidates():
    from itertools import product
    params = []
    for th, n_cnt in product((2, 3, 6, 12, 24), (2, 3, 4)):
        params.append({"max_child": th,
                       "cut_depth": n_cnt})
    return params


def init_ltgen_with_params(conf, table, params, **_):
    max_child = params["max_child"]
    cut_depth = params["cut_depth"]
    type_func_name = conf.get("log_template_fttree", "type_func")
    if type_func_name == "top":
        message_type_func = LTGenFTTree.message_type_top
    elif type_func_name == "length":
        message_type_func = LTGenFTTree.message_type_length
    else:
        message_type_func = None
    return LTGenFTTree(table, max_child, cut_depth, message_type_func)




#!/usr/bin/env python
# coding: utf-8

import numpy as np
from scipy.special import comb


def _comb2(n):
    return comb(n, 2, exact=True)


def confusion_matrix(labels_true, labels_pred):
    from sklearn.metrics.cluster import contingency_matrix

    a_true = np.array(labels_true)
    a_pred = np.array(labels_pred)
    cm = contingency_matrix(a_true, a_pred, sparse=True)

    nz_true, nz_pred = cm.nonzero()
    nz_cnt = cm.data

    tp = sum(_comb2(n_ij) for n_ij in cm.data)

    # false negative: same cluster in labels_true, but different in labels_pred
    fn = 0
    for uniq_label, uniq_cnt in zip(*np.unique(nz_true, return_counts=True)):
        if uniq_cnt > 1:
            childs = nz_cnt[nz_true == uniq_label]
            # add combinations except true_positive part
            fn += _comb2(sum(childs)) - sum(_comb2(c) for c in childs)

    # false positive: different cluster in labels_true, but same in labels_pred
    fp = 0
    for uniq_label, uniq_cnt in zip(*np.unique(nz_pred, return_counts=True)):
        if uniq_cnt > 1:
            childs = nz_cnt[nz_pred == uniq_label]
            # add combinations except true_positive part
            fp += _comb2(sum(childs)) - sum(_comb2(c) for c in childs)

    total = _comb2(a_true.shape[0])
    tn = total - tp - fn - fp

    return np.array([[tp, fp], [fn, tn]])


def rand_score(labels_true, labels_pred):
    tp, fp, fn, tn = confusion_matrix(labels_true, labels_pred).ravel()
    return (tp + tn) / (tp + fp + fn + tn)


def precision_recall_fscore(labels_true, labels_pred):
    tp, fp, fn, tn = confusion_matrix(labels_true, labels_pred).ravel()
    precision = tp / (tp + fp)
    recall = tp / (tp + fn)
    fscore = (2. * precision * recall) / (precision + recall)
    return precision, recall, fscore


def parsing_accuracy(labels_true, labels_pred):
    """Parsing accuracy is one of clustering metrics.
    It is the ratio of log lines in completely-correct clusters.
    A completely-correct cluster has log instances (i.e., cluster members)
    that is completely same as the instances in an answer cluster.

    This metrics is defined in [1].
    [1] P. He, et al. Drain: An Online Log Parsing Approach with Fixed Depth Tree. ICWS 2017, pp.33–40, 2017.
    https://github.com/logpai/logparser is also referred.
    """

    from sklearn.metrics.cluster import contingency_matrix
    a_true = np.array(labels_true)
    a_pred = np.array(labels_pred)
    n_total_line = a_true.shape[0]
    cm = contingency_matrix(a_true, a_pred, sparse=True)

    nz_true, nz_pred = cm.nonzero()
    nz_cnt = cm.data

    n_correct_line = 0
    for uniq_label_true, uniq_cnt_true in zip(*np.unique(nz_true, return_counts=True)):
        # 1 estimated cluster for 1 answer cluster?
        if uniq_cnt_true == 1:
            index_uniq_true = (nz_true == uniq_label_true)
            index_uniq_pred = (nz_pred == nz_pred[index_uniq_true][0])
            # 1 answer cluster for 1 estimated cluster?
            if nz_true[index_uniq_pred].shape[0] == 1:
                n_correct_line += sum(nz_cnt[index_uniq_true])

    return 1. * n_correct_line / n_total_line


def cluster_accuracy(labels_true, labels_pred):
    """Cluster accuracy is one of clustering metrics.
    This is an extended version of parsing accuracy,
    but counting clusters instead of log messages.
    """
    from sklearn.metrics.cluster import contingency_matrix
    a_true = np.array(labels_true)
    a_pred = np.array(labels_pred)
    n_cluster = np.unique(a_true).shape[0]
    cm = contingency_matrix(a_true, a_pred, sparse=True)

    nz_true, nz_pred = cm.nonzero()

    n_correct_cluster = 0
    for uniq_label_true, uniq_cnt_true in zip(*np.unique(nz_true, return_counts=True)):
        # 1 estimated cluster for 1 answer cluster?
        if uniq_cnt_true == 1:
            index_uniq_true = (nz_true == uniq_label_true)
            index_uniq_pred = (nz_pred == nz_pred[index_uniq_true][0])
            # 1 answer cluster for 1 estimated cluster?
            if nz_true[index_uniq_pred].shape[0] == 1:
                n_correct_cluster += 1

    return 1. * n_correct_cluster / n_cluster


def over_division_cluster_ratio(labels_true, labels_pred):
    from sklearn.metrics.cluster import contingency_matrix
    a_true = np.array(labels_true)
    a_pred = np.array(labels_pred)
    n_cluster = np.unique(a_true).shape[0]
    cm = contingency_matrix(a_true, a_pred, sparse=True)

    nz_true, nz_pred = cm.nonzero()

    n_fail_cluster = 0
    for uniq_label_true, uniq_cnt_true in zip(*np.unique(nz_true, return_counts=True)):
        # multiple estimated cluster for 1 answer cluster?
        if uniq_cnt_true > 1:
            n_fail_cluster += 1

    return 1. * n_fail_cluster / n_cluster


def over_aggregation_cluster_ratio(labels_true, labels_pred):
    from sklearn.metrics.cluster import contingency_matrix
    a_true = np.array(labels_true)
    a_pred = np.array(labels_pred)
    n_cluster = np.unique(a_true).shape[0]
    cm = contingency_matrix(a_true, a_pred, sparse=True)

    nz_true, nz_pred = cm.nonzero()

    n_fail_cluster = 0
    for uniq_label_true in np.unique(nz_true):
        index_uniq_true = (nz_true == uniq_label_true)
        index_uniq_pred = (nz_pred == nz_pred[index_uniq_true][0])
        # multiple answer cluster for 1 estimated cluster?
        if nz_true[index_uniq_pred].shape[0] > 1:
            n_fail_cluster += 1

    return 1. * n_fail_cluster / n_cluster






#!/usr/bin/env python
# coding: utf-8

import os
import unittest
import tempfile

from amulog import config
from amulog import log_db
from amulog import manager

from amulog import testutil


class TestMysql(unittest.TestCase):

    _path_testlog = None
    _path_testdb = None
    _path_ltgendump = None

    @classmethod
    def setUpClass(cls):
        fd_testlog, cls._path_testlog = tempfile.mkstemp()
        os.close(fd_testlog)
        fd_ltgendump, cls._path_ltgendump = tempfile.mkstemp()
        os.close(fd_ltgendump)

        tlg = testutil.TestLogGenerator(testutil.DEFAULT_CONFIG, seed=3)
        tlg.dump_log(cls._path_testlog)

    @classmethod
    def tearDownClass(cls):
        os.remove(cls._path_testlog)
        os.remove(cls._path_ltgendump)

    @classmethod
    def _conf_mysql(cls):
        conf = config.open_config()
        conf['general']['src_path'] = cls._path_testlog
        conf['database']['database'] = "mysql"
        conf['database']['mysql_host'] = "localhost"
        conf['database']['mysql_dbname'] = "test_amulog"
        conf['database']['mysql_user'] = "testamulog"
        conf['database']['mysql_passwd'] = "testamulog"
        conf['manager']['indata_filename'] = cls._path_ltgendump
        return conf

    def test_makedb_online(self):
        conf = self._conf_mysql()
        from amulog import __main__ as amulog_main
        targets = amulog_main.get_targets_conf(conf)
        manager.process_files_online(conf, targets, reset_db=True)

        ld = log_db.LogData(conf)
        num = ld.count_lines()
        self.assertEqual(num, 6539,
                         "not all logs added to database")
        ltg_num = len([gid for gid in ld.iter_ltgid()])
        self.assertTrue(3 < ltg_num < 20,
                        ("log template generation fails? "
                         "(groups: {0})".format(ltg_num)))
        ld.drop_all()

    def test_makedb_offline(self):
        conf = self._conf_mysql()
        from amulog import __main__ as amulog_main
        targets = amulog_main.get_targets_conf(conf)
        manager.process_files_offline(conf, targets, reset_db=True)

        ld = log_db.LogData(conf)
        num = ld.count_lines()
        self.assertEqual(num, 6539,
                         "not all logs added to database")
        ltg_num = len([gid for gid in ld.iter_ltgid()])
        self.assertTrue(3 < ltg_num < 20,
                        ("log template generation fails? "
                         "(groups: {0})".format(ltg_num)))
        ld.drop_all()



#!/usr/bin/env python
# coding: utf-8

import sys
import os
import datetime
import logging
import configparser
from collections import defaultdict
from dateutil.tz import tzlocal

CONFIG_ENV = "AMULOG_CONFIG"
DEFAULT_CONFIG = "/".join((os.path.dirname(os.path.abspath(__file__)),
                           "data/config.conf.default"))
LOAD_SECTION = 'general'
LOAD_OPTION = 'base_filename'
IMPORT_SECTION = 'general'
IMPORT_OPTION = 'import'


class GroupDef:
    """
    Define grouping by external text
    Rules:
        description after # in a line will be recognized as comment
        line "[GROUP_NAME]" will change group to set
        other lines add elements in the group set with GROUP_NAME line
    """

    def __init__(self, fn, default_val=None):
        self.gdict = {}
        self.rgdict = {}
        self.default = default_val
        if fn is None or fn == "":
            pass
        else:
            self.open_def(fn)

    def open_def(self, fn):
        group = None
        with open(fn, 'r') as f:
            for line in f:
                # ignore after comment sygnal
                line = line.strip().partition("#")[0]
                if line == "":
                    continue
                elif line[0] == "#":
                    continue
                elif line[0] == "[" and line[-1] == "]":
                    group = line[1:].strip("[]")
                else:
                    if group is None:
                        raise ValueError("no group definition before value")
                    val = line
                    self.gdict.setdefault(group, []).append(val)
                    self.rgdict.setdefault(val, []).append(group)

    def setdefault(self, group):
        self.default = group

    def groups(self):
        return self.gdict.keys()

    def values(self):
        return self.rgdict.keys()

    def ingroup(self, group, val):
        return val in self.gdict[group]

    def get_group(self, val):
        if val in self.rgdict:
            return self.rgdict[val]
        else:
            return []

    def get_value(self, group):
        if group in self.gdict:
            return self.gdict[group]
        else:
            return []

    def iter_def(self):
        for group, l_val in self.gdict.items():
            for val in l_val:
                yield group, val


def gettuple(conf, section, name, sep=","):
    ret = conf.get(section, name)
    if ret.strip() == "":
        return tuple()
    else:
        return tuple(e.strip() for e in ret.split(sep)
                     if not len(e.strip()) == 0)


def getlist(conf, section, name, sep=","):
    ret = conf.get(section, name)
    if ret.strip() == "":
        return []
    else:
        return [e.strip() for e in ret.split(sep)
                if not len(e.strip()) == 0]


def getdict(conf, section, name):
    val = conf.get(section, name)
    ret = {}
    for e in val.split(","):
        e = e.strip()
        assert "=" in e
        k, _, v = e.partition("=")
        ret[k] = v
    return ret


def getdt(conf, section, name):
    ret = conf.get(section, name)
    if ret == "":
        return None
    else:
        dt = datetime.datetime.strptime(ret.strip(), "%Y-%m-%d %H:%M:%S")
        dt = dt.replace(tzinfo=tzlocal())
        return dt


def getterm(conf, section, name):
    val = conf.get(section, name)
    if val == "":
        return None
    else:
        ret = []
        for e in val.split(","):
            dt = datetime.datetime.strptime(e.strip(), "%Y-%m-%d %H:%M:%S")
            dt = dt.replace(tzinfo=tzlocal())
            ret.append(dt)
        return ret


def getdur(conf, section, name):
    ret = conf.get(section, name)
    if ret == "":
        return None
    else:
        return str2dur(ret)


def getname(conf):
    return conf.get(LOAD_SECTION, LOAD_OPTION)


def str2dur(string):
    """
    Note:
        \\d+s: \\d seconds
        \\d+m: \\d minutes
        \\d+h: \\d hours
        \\d+d: \\d days
        \\d+w: \\d * 7 days
    """
    if "s" in string:
        num = int(string.partition("s")[0])
        return datetime.timedelta(seconds=num)
    elif "m" in string:
        num = int(string.partition("m")[0])
        return datetime.timedelta(minutes=num)
    elif "h" in string:
        num = int(string.partition("h")[0])
        return datetime.timedelta(hours=num)
    elif "d" in string:
        num = int(string.partition("d")[0])
        return datetime.timedelta(days=num)
    elif "w" in string:
        num = int(string.partition("w")[0])
        return datetime.timedelta(days=num * 7)
    else:
        raise ValueError("Duration string invalid")


def dur2str(td):
    samples = [("m", 60), ("h", 60), ("d", 24), ("w", 7)]

    footer = "s"
    val = int(td.total_seconds())
    for tmp_footer, mod in samples:
        if val % mod == 0:
            footer = tmp_footer
            val = val // mod
        else:
            break
    return str(val) + footer


def merge_config(conf1, conf2):
    """Merge conf1 with conf2.
    If both config has same options, conf2 has priority."""
    for sec in conf2.sections():
        for opt in conf2.options(sec):
            if not conf1.has_section(sec):
                conf1.add_section(sec)
            conf1.set(sec, opt, conf2[sec][opt])
    return conf1


def load_defaults(iterable_conf_path=None):
    l_fn = iterable_conf_path
    base_conf = None
    for fn in l_fn:
        if not os.path.exists(fn):
            raise IOError("{0} not found".format(fn))
        tmp_conf = configparser.ConfigParser()
        tmp_conf.read(fn)
        if base_conf is None:
            base_conf = tmp_conf
        else:
            base_conf = merge_config(base_conf, tmp_conf)

    return base_conf


def _load_imports(conf):
    while conf.has_option(IMPORT_SECTION, IMPORT_OPTION):
        if conf[IMPORT_SECTION][IMPORT_OPTION] == "":
            break
        import_fn = conf[IMPORT_SECTION][IMPORT_OPTION]
        conf.remove_option(IMPORT_SECTION, IMPORT_OPTION)
        if not os.path.exists(import_fn):
            raise IOError("config load error ({0})".format(import_fn))
        import_conf = configparser.ConfigParser()
        import_conf.read(import_fn)
        conf = merge_config(import_conf, conf)

    return conf


def open_config(fn=None, env=CONFIG_ENV,
                ex_defaults=None,
                base_default=True, ignore_import=False,
                verbose=True):
    """
    Args:
        fn (str, optional): Configuration file path.
        env (str, optional): If fn is None, use Environment Variable of given name.
        ex_defaults (List[str], optional): External default configurations.
                                 If you use other packages with amulog,
                                 use this to add its configurations.
        base_default (bool, optional): Use amulog default values for missing options.
        ignore_import (bool, optional): Ignore general.import.
        verbose (bool, optional): Notify if no available config except default one.
        
    """

    conf = configparser.ConfigParser()

    if fn is None and env is not None:
        fn = os.environ.get(env)

    if fn is None:
        if verbose:
            print("Processing with default configuration ...")
    else:
        if not os.path.exists(fn):
            raise IOError("{0} not found".format(fn))
        ret = conf.read(fn)
        if len(ret) == 0:
            raise IOError("config load error ({0})".format(fn))
        conf.set(LOAD_SECTION, LOAD_OPTION, fn)

    if not ignore_import:
        conf = _load_imports(conf)

    l_defaults = []
    if base_default:
        l_defaults.append(DEFAULT_CONFIG)
    if ex_defaults:
        l_defaults += ex_defaults
    if l_defaults:
        default_conf = load_defaults(l_defaults)
        conf = merge_config(default_conf, conf)

    return conf


def show_config(conf):
    for section in conf.sections():
        print("[{0}]".format(section))
        for option in conf.options(section):
            print("{0} = {1}".format(option, conf[section][option]))
        print()


def show_default_config(ex_defaults=None):
    conf = load_defaults(ex_defaults)
    show_config(conf)


def show_config_diff(l_conf_name, l_conf=None, ex_defaults=None):
    from . import common
    if l_conf is None:
        l_conf = [open_config(conf_path, ex_defaults)
                  for conf_path in l_conf_name]

    def iter_secopt(_conf):
        for _sec in _conf.sections():
            for _opt in _conf.options(_sec):
                yield _sec, _opt

    keys = set()
    for conf in l_conf:
        keys = keys | set(iter_secopt(conf))

    d = defaultdict(lambda: defaultdict(set))
    for conf in l_conf:
        for key in keys:
            sec, opt = key
            if conf.has_option(sec, opt):
                d[sec][opt].add(conf[sec][opt])

    d_keys = defaultdict(list)
    for sec in d:
        for opt in d[sec]:
            if len(d[sec][opt]) > 1:
                d_keys[sec].append(opt)

    for sec, l_opt in d_keys.items():
        print("[{0}]".format(sec))
        for opt in l_opt:
            print("{0} = ...".format(opt))
            buf = []
            for name, conf in zip(l_conf_name, l_conf):
                if conf.has_option(sec, opt):
                    buf.append([name, conf[sec][opt]])
            print(common.cli_table(buf, spl=": "))
        print()


def write(name, conf):
    with open(name, "w") as f:
        conf.write(f)


def minimize(conf, ex_defaults=None, ignore_import=False):
    # def add_opt(conf, sec, opt, val):
    #     if sec not in conf:
    #         conf[sec] = {}
    #     conf[sec][opt] = val

    new_conf = configparser.ConfigParser()
    default_conf = open_config(None, ex_defaults)
    if not ignore_import:
        conf = _load_imports(conf)

    # import_conf = configparser.ConfigParser()
    # if conf.has_option(IMPORT_SECTION, IMPORT_OPTION):
    #     import_fn = conf[IMPORT_SECTION][IMPORT_OPTION]
    #     if os.path.exists(import_fn):
    #         import_conf.read(import_fn)

    for sec in conf.sections():
        for opt in conf.options(sec):
            # if import_conf.has_option(sec, opt):
            #     if import_conf[sec][opt] == conf[sec][opt]:
            #         continue

            if not default_conf.has_option(sec, opt):
                # undefine key in defaults
                if sec not in new_conf:
                    new_conf[sec] = {}
                new_conf[sec][opt] = conf[sec][opt]
            elif conf[sec][opt] == default_conf[sec][opt]:
                # same value from defaults
                pass
            else:
                # different value from defaults
                if sec not in new_conf:
                    new_conf[sec] = {}
                new_conf[sec][opt] = conf[sec][opt]

    return new_conf


def config_minimum(fn, ex_defaults=None):
    conf = open_config(fn, ex_defaults)
    default_conf = open_config(None)

    for sec in conf.sections():
        l_diff = []
        for opt in conf.options(sec):
            if not default_conf.has_option(sec, opt):
                l_diff.append((opt, conf[sec][opt]))
            elif conf[sec][opt] == default_conf[sec][opt]:
                pass
            else:
                l_diff.append((opt, conf[sec][opt]))
        if len(l_diff) > 0:
            print("[{0}]".format(sec))
            for opt, val in l_diff:
                print("{0} = {1}".format(opt, val))
            print()


def config_group_edit(l_conf_name, d_rule, l_conf=None):
    if l_conf is None:
        l_conf = [open_config(conf_path, base_default=False, ignore_import=True)
                  for conf_path in l_conf_name]
    for key, l_val in d_rule.items():
        if isinstance(key, str):
            sec, opt = key.split(".")
        elif len(key) == 2:
            sec, opt = key
        else:
            raise ValueError
        assert len(l_val) == len(l_conf_name), "bad value length"
        for conf, val in zip(l_conf, l_val):
            if not conf.has_section(sec):
                conf[sec] = {}
            conf[sec][opt] = str(val)

    for name, conf in zip(l_conf_name, l_conf):
        write(name, conf)


def config_shadow(n=1, cond=None, incr=None, fn=None, output=None,
                  ignore_overwrite=False, ex_defaults=None):
    if cond is None:
        cond = {}
    if incr is None:
        incr = []

    l_ret = []
    for i in range(n):
        conf = open_config(fn, ex_defaults, ignore_import=True)
        for key, val in cond.items():
            sec, opt = key.split(".")
            conf[sec][opt] = val
        for key in incr:
            sec, opt = key.split(".")
            conf[sec][opt] = conf[sec][opt] + str(i)

        if n == 1:
            footer = ""
        else:
            footer = str(i)
        if output is None:
            if fn is None:
                temp_output = "copy.conf" + footer
            else:
                temp_output = fn + str(i)
        else:
            temp_output = output + footer

        if os.path.exists(temp_output) and not ignore_overwrite:
            raise IOError("{0} already exists, use -f to ignore".format(
                temp_output))

        write(temp_output, minimize(conf, ex_defaults, ignore_import=True))
        l_ret.append(temp_output)
        print("{0}".format(temp_output))
    return l_ret


def check_all_diff(l_conf_name, keys, l_conf=None):
    """Return True if all configs have different values in given keys."""
    if l_conf is None:
        l_conf = [open_config(conf_name) for conf_name in l_conf_name]
    ret = True
    for key in keys:
        if isinstance(key, tuple) or isinstance(key, list):
            sec, opt = key
        else:
            sec, opt = key.split(".")
        values = [conf[sec][opt] for conf in l_conf]
        if len(values) == len(set(values)):
            pass
        else:
            print("Notice: Same {0} settings found".format(key))
            for name, conf in zip(l_conf_name, l_conf):
                print("{0} : {1}".format(name, conf[sec][opt]))
            print()
            ret = False
    return ret


def read_config_group(cgroup_path):
    ret = []
    with open(cgroup_path, "r") as f:
        for line in f:
            fp = line.strip()
            if fp == "":
                pass
            else:
                ret.append(fp)
    return ret


def load_config_group(cgroup_path, ex_defaults=None):
    l_conf = []
    for fp in read_config_group(cgroup_path):
        if os.path.exists(fp):
            l_conf.append(open_config(fp, ex_defaults))
        else:
            sys.stderr.write(
                "Warning: Invalid config name {0}".format(fp))
    return l_conf


def dump_config_group(cgroup_name, l_conf_name):
    with open(cgroup_name, "w") as f:
        f.write("\n".join(l_conf_name))


def set_logging_stdio(logger=None, logger_name=None,
                      lv=logging.INFO):
    fmt = logging.Formatter(
        fmt="%(asctime)s %(levelname)s (%(processName)s) %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S")
    ch = logging.StreamHandler()
    ch.setFormatter(fmt)
    ch.setLevel(lv)

    temp_loggers = []
    if logger is None:
        pass
    elif isinstance(logger, list):
        temp_loggers += logger
    elif isinstance(logger, logging.Logger):
        temp_loggers.append(logger)
    else:
        raise TypeError
    if logger_name is None:
        pass
    elif isinstance(logger_name, list):
        temp_loggers += [logging.getLogger(ln) for ln in logger_name]
    elif isinstance(logger_name, str):
        temp_loggers.append(logging.getLogger(logger_name))
    else:
        raise TypeError

    for logger in temp_loggers:
        logger.setLevel(lv)
        logger.addHandler(ch)
    return ch


# common objects for logging
def set_common_logging(conf, logger=None, logger_name=None,
                       lv=logging.INFO):
    """
    Args:
        conf
        logger (logging.Logger or list[logging.Logger])
        logger_name (str or list[str])
        lv (int): logging level
    Returns:
        logging.SomeHandler
    """
    fn = conf.get("general", "logging")
    fmt = logging.Formatter(
        fmt="%(asctime)s %(levelname)s (%(processName)s) %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S")
    # lv = logging.INFO
    if fn == "":
        ch = logging.StreamHandler()
    else:
        ch = logging.FileHandler(fn)
    ch.setFormatter(fmt)
    ch.setLevel(lv)

    temp_loggers = []
    if logger is None:
        pass
    elif isinstance(logger, list):
        temp_loggers += logger
    elif isinstance(logger, logging.Logger):
        temp_loggers.append(logger)
    else:
        raise TypeError
    if logger_name is None:
        pass
    elif isinstance(logger_name, list):
        temp_loggers += [logging.getLogger(ln) for ln in logger_name]
    elif isinstance(logger_name, str):
        temp_loggers.append(logging.getLogger(logger_name))
    else:
        raise TypeError

    for logger in temp_loggers:
        logger.setLevel(lv)
        logger.addHandler(ch)
        if lv <= logging.DEBUG:
            logger.debug("logger is on debug mode")

    return ch


def release_common_logging(ch, logger=None, logger_name=None):
    temp_loggers = []
    if logger is None:
        pass
    elif isinstance(logger, list):
        temp_loggers += logger
    elif isinstance(logger, logging.Logger):
        temp_loggers.append(logger)
    else:
        raise TypeError
    if logger_name is None:
        pass
    elif isinstance(logger, list):
        temp_loggers += [logging.getLogger(ln) for ln in logger_name]
    elif isinstance(logger_name, str):
        temp_loggers.append(logging.getLogger(logger_name))
    else:
        raise TypeError

    for logger in temp_loggers:
        logger.removeHandler(ch)



#!/usr/bin/env python
# coding: utf-8

import os
import unittest
import tempfile

from amulog import common
from amulog import config
from amulog import manager
from amulog import lt_common

from amulog import testutil


class TestLTGen(unittest.TestCase):

    _path_testlog = None

    @classmethod
    def setUpClass(cls):
        fd_testlog, cls._path_testlog = tempfile.mkstemp()
        os.close(fd_testlog)
        tlg = testutil.TestLogGenerator(testutil.DEFAULT_CONFIG, seed=3)
        tlg.dump_log(cls._path_testlog)

    @classmethod
    def tearDownClass(cls):
        os.remove(cls._path_testlog)

    def _try_method(self, conf, online=True):
        table = lt_common.TemplateTable()
        ltgen = manager.init_ltgen_methods(conf, table)

        iterobj = manager.iter_plines(conf, [self._path_testlog])
        if online:
            for pline in iterobj:
                ltgen.process_line(pline)
        else:
            d_pline = {mid: pline for mid, pline in enumerate(iterobj)}
            ltgen.process_offline(d_pline)
        return table

    def test_import(self):
        conf = config.open_config(verbose=False)
        conf['log_template']['lt_methods'] = "import"
        tpl_path = common.filepath_local(__file__, "testlog_tpl.txt")
        conf['log_template_import']['def_path'] = tpl_path
        table = self._try_method(conf)

        n_tpl = len(table)
        self.assertTrue(n_tpl == 6)

    def test_drain(self):
        conf = config.open_config(verbose=False)
        conf['log_template']['lt_methods'] = "drain"
        table = self._try_method(conf)

        n_tpl = len(table)
        self.assertTrue(3 < n_tpl < 20)

    def test_lenma(self):
        conf = config.open_config(verbose=False)
        conf['log_template']['lt_methods'] = "lenma"
        table = self._try_method(conf)

        n_tpl = len(table)
        self.assertTrue(3 < n_tpl < 20)

    def test_dlog(self):
        conf = config.open_config(verbose=False)
        conf['log_template']['lt_methods'] = "dlog"
        table = self._try_method(conf, online=False)

        n_tpl = len(table)
        self.assertTrue(3 < n_tpl < 300)

    def test_fttree(self):
        conf = config.open_config(verbose=False)
        conf['log_template']['lt_methods'] = "fttree"
        table = self._try_method(conf)

        n_tpl = len(table)
        self.assertTrue(3 < n_tpl < 50)

    def test_va(self):
        conf = config.open_config(verbose=False)
        conf['log_template']['lt_methods'] = "va"
        table = self._try_method(conf)

        n_tpl = len(table)
        self.assertTrue(3 < n_tpl < 20)

    def test_shiso_first(self):
        conf = config.open_config(verbose=False)
        conf['log_template']['lt_methods'] = "shiso"
        table = self._try_method(conf)

        n_tpl = len(table)
        self.assertTrue(3 < n_tpl < 20)


if __name__ == "__main__":
    unittest.main()





"""
SHISO: a log template generation algorithm proposed in [1].
[1] Masayoshi Mizutani. Incremental Mining of System Log Format.
    in IEEE 10th International Conference on Services Computing, pp. 595–602, 2013.
"""

import logging
import numpy as np

from amulog import lt_common
from amulog import lt_misc

_logger = logging.getLogger(__package__)


class LTGenNode:

    def __init__(self, tid=None):
        self.l_child = []
        self.tid = tid

    def __len__(self):
        return len(self.l_child)

    def __iter__(self):
        return self.l_child.__iter__()

    def join(self, node):
        self.l_child.append(node)


class LTGenSHISO(lt_common.LTGen):

    def __init__(self, table, threshold, max_child, cfunc=None):
        super(LTGenSHISO, self).__init__(table)
        self._n_root = LTGenNode()
        self._threshold = threshold
        self._max_child = max_child
        if cfunc is None:
            self._cfunc = self.c_alphabet

    def load(self, loadobj):
        self._n_root = loadobj

    def dumpobj(self):
        return self._n_root

    def process_line(self, pline):
        l_w = pline["words"]
        n_parent = self._n_root
        while True:
            for n_child in n_parent:
                _logger.debug(
                    "comparing with tid {0}".format(n_child.tid))
                nc_tpl = self._table[n_child.tid]
                sr = self._seq_ratio(nc_tpl, l_w, self._cfunc)
                _logger.debug("seq_ratio : {0}".format(sr))

                if sr >= self._threshold:
                    # add input message into the node cluster
                    _logger.debug(
                        "merged with tid {0}".format(n_child.tid))
                    state = self.merge_tpl(l_w, n_child.tid)
                    return n_child.tid, state
            else:
                if len(n_parent) < self._max_child:
                    # add new node
                    _logger.debug("no node to be merged, add new node")
                    new_tid = self.add_tpl(l_w)
                    n_child = LTGenNode(new_tid)
                    n_parent.join(n_child)
                    return new_tid, self.state_added
                else:
                    # select child node
                    a_sim = np.array([self._sim(self.get_tpl(n_child.tid), l_w)
                                      for n_child in n_parent])
                    n_parent = n_parent.l_child[a_sim.argmin()]

    @staticmethod
    def c_original(w):
        a_cnt = np.zeros(4)  # upper_alphabet, lower_alphabet, digit, symbol
        for c in w:
            if c.isupper():
                a_cnt[0] += 1
            elif c.islower():
                a_cnt[1] += 1
            elif c.isdigit():
                a_cnt[2] += 1
            else:
                a_cnt[3] += 1
        return a_cnt / np.linalg.norm(a_cnt)

    @staticmethod
    def c_alphabet(w):
        a_cnt = np.zeros(26 + 26 + 2)  # A-Z, a-z, digit, symbol
        for c in w:
            if c.isupper():
                abc_index = ord(c) - 65
                a_cnt[abc_index] += 1.0
            elif c.islower():
                abc_index = ord(c) - 97
                a_cnt[26 + abc_index] += 1
            elif c.isdigit():
                a_cnt[-2] += 1.0
            else:
                a_cnt[-1] += 1.0
        return a_cnt / np.linalg.norm(a_cnt)

    @staticmethod
    def _seq_ratio(m1, m2, cfunc):
        if len(m1) == len(m2):
            length = len(m1)
            if length == 0:
                return 1.0

            sum_dist = 0.0
            for w1, w2 in zip(m1, m2):
                if lt_common.REPLACER in (w1, w2):
                    pass
                else:
                    sum_dist += sum(np.power(cfunc(w1) - cfunc(w2), 2))
            return 1.0 - (sum_dist / (2.0 * length))
        else:
            return 0.0

    @staticmethod
    def _sim(m1, m2):
        """Sim is different from SeqRatio, because Sim allows
        word sequence of different length. Original SHISO uses
        Levenshtein edit distance, and amulog follows that."""
        return lt_misc.edit_distance(m1, m2)

    @staticmethod
    def _equal(m1, m2):
        if not len(m1) == len(m2):
            return False
        for w1, w2 in zip(m1, m2):
            if w1 == w2 or w1 == lt_common.REPLACER or w2 == lt_common.REPLACER:
                pass
            else:
                return False
        else:
            return True


class LTGroupSHISO(lt_common.LTGroupOnline):

    def __init__(self, lttable, ngram_length=3,
                 th_lookup=0.3, th_distance=0.85, mem_ngram=True):
        super(LTGroupSHISO, self).__init__()
        self._lttable = lttable
        # self.d_group = {} # key : groupid, val : [ltline, ...]
        # self.d_rgroup = {} # key : ltid, val : groupid
        self.ngram_length = ngram_length
        self.th_lookup = th_lookup
        self.th_distance = th_distance
        self.mem_ngram = mem_ngram
        if self.mem_ngram:
            self.d_ngram = {}  # key : ltid, val : [str, ...]

    def add(self, lt_new):
        _logger.debug("group search for ltid {0}".format(lt_new.ltid))

        r_max = 0.0
        lt_max = None
        l_cnt = [0] * len(self._lttable)
        l_ng1 = self._get_ngram(lt_new)
        for ng in l_ng1:
            for lt_temp, l_ng2 in self._lookup_ngram(ng):
                l_cnt[lt_temp.ltid] += 1
                r = 2.0 * l_cnt[lt_temp.ltid] / (len(l_ng1) + len(l_ng2))
                if r > r_max:
                    r_max = r
                    lt_max = lt_temp
        _logger.debug("ngram co-occurrance map : {0}".format(l_cnt))
        _logger.debug("r_max : {0}".format(r_max))
        if r_max > self.th_lookup:
            assert lt_max is not None, "bad threshold for lt group lookup"
            _logger.debug("lt_max ltid : {0}".format(lt_max.ltid))
            d = 2.0 * lt_misc.edit_distance(lt_new.ltw, lt_max.ltw) / \
                (len(lt_new.ltw) + len(lt_max.ltw))
            _logger.debug("edit distance ratio : {0}".format(d))
            if d < self.th_distance:
                gid = self._mk_group(lt_new, lt_max)
                _logger.debug("smaller than threshold")
                _logger.debug("merge it (gid {0})".format(gid))
                _logger.debug("gid {0} : {1}".format(gid,
                                                     [ltline.ltid for ltline in self._d_group[gid]]))
                return gid
        # No merge, make group with single lt
        gid = self._mk_group_single(lt_new)
        _logger.debug("No similar format found, make group with single lt")
        _logger.debug("gid {0} : {1}".format(gid,
                                             [ltline.ltid for ltline in self._d_group[gid]]))
        return gid

    def _mk_group(self, lt_new, lt_old):
        assert lt_new.ltid not in self._d_rgroup
        if lt_old.ltid in self._d_rgroup:
            groupid = self._d_rgroup[lt_old.ltid]
            self.add_lt(groupid, lt_new)
        else:
            groupid = self._next_groupid()
            self.add_lt(groupid, lt_old)
            self.add_lt(groupid, lt_new)
        return groupid

    def _mk_group_single(self, lt_new):
        assert lt_new.ltid not in self._d_rgroup
        groupid = self._next_groupid()
        self.add_lt(groupid, lt_new)
        return groupid

    def _get_ngram(self, ltline):

        def ngram(ltw, length):
            return [ltw[i:i + length] for i in range(len(ltw) - length)]

        if self.mem_ngram:
            if ltline.ltid not in self.d_ngram:
                self.d_ngram[ltline.ltid] = \
                    ngram(ltline.ltw, self.ngram_length)
            return self.d_ngram[ltline.ltid]
        else:
            return ngram(ltline.ltw, self.ngram_length)

    def _lookup_ngram(self, ng):
        ret = []
        for ltline in self._lttable:
            if ng in self._get_ngram(ltline):
                ret.append((ltline, ng))
        return ret

    def load(self, obj):
        pass

    def dumpobj(self):
        return None


def init_ltgen(conf, table, **_):
    threshold = conf.getfloat("log_template_shiso", "ltgen_threshold")
    max_child = conf.getint("log_template_shiso", "ltgen_max_child")
    return LTGenSHISO(table, threshold, max_child)



import pymysql

from . import db_common

pymysql.install_as_MySQLdb()


class Mysql(db_common.Database):

    def __init__(self, host, dbname, user, passwd):
        self._host = host
        self._dbname = dbname
        self._user = user
        self._passwd = passwd
        self._connect = None

    def __del__(self):
        if self._connect is not None:
            self._connect.commit()
            self._connect.close()

    def _open(self):
        if not self.db_exists():
            self._init_database()
        self._connect = pymysql.connect(host=self._host, db=self._dbname,
                                        user=self._user, passwd=self._passwd)

    def _init_database(self):
        connect = self._connect_root()
        cursor = connect.cursor()
        cursor.execute("create database {0}".format(self._dbname))

    def _connect_root(self):
        return pymysql.connect(host=self._host, user=self._user,
                               passwd=self._passwd)

    def db_exists(self):
        connect = self._connect_root()
        cursor = connect.cursor()
        cursor.execute("show databases")
        return self._dbname in [row[0] for row in cursor]

    def reset(self):
        connect = self._connect_root()
        cursor = connect.cursor()
        sql = "drop database if exists {0}".format(self._dbname)
        cursor.execute(sql)
        connect.commit()

    def commit(self):
        if self._connect is not None:
            self._connect.commit()

    def is_internal_table(self, name):
        return False

    @classmethod
    def datetime(cls, ret):
        return ret

    @staticmethod
    def _ph(varname):
        return "%({0})s".format(varname)

    @staticmethod
    def _table_key_type(type_str):
        if type_str == "integer":
            return "int"
        else:
            return type_str

    @staticmethod
    def _table_key_attr(attr):
        if attr == "primary_key":
            return "primary key"
        elif attr == "auto_increment":
            # differ from sqlite
            return "auto_increment"
        elif attr == "not_null":
            return "not null"
        else:
            raise NotImplementedError

    @staticmethod
    def _index_key(tablekey):
        if tablekey.type == "text":
            return "{0}({1})".format(tablekey.key, tablekey.attr[0])
        else:
            return tablekey.key

    def execute(self, sql, args=None):
        # print(sql)
        # if args is not None and len(args) > 0:
        #     print(args)
        if self._connect is None:
            self._open()
        cursor = self._connect.cursor()
        if args is not None and len(args) == 0:
            cursor.execute(sql)
        else:
            cursor.execute(sql, args)
        return cursor

    def executemany(self, sql, iter_args):
        if self._connect is None:
            self._open()
        cursor = self._connect.cursor()
        cursor.executemany(sql, iter_args)
        return cursor

    def get_table_names(self):
        sql = "show tables"
        cursor = self.execute(sql)
        return [row[0] for row in cursor]

    def get_column_names(self, table_name):
        raise NotImplementedError


def init_db_conn(conf):
    host = conf.get("database", "mysql_host")
    dbname = conf.get("database", "mysql_dbname")
    user = conf.get("database", "mysql_user")
    passwd = conf.get("database", "mysql_passwd")
    return Mysql(host, dbname, user, passwd)

